import rustworkx as rx
import copy
from functools import lru_cache
from typing import List, TypeVar, Optional, Any, Union, Dict, Iterator, Tuple
from types import MappingProxyType
import numpy as np

T = TypeVar('T')


class Graph:
    def __init__(self, directed: bool = False):
        """Initialize an empty Klotho graph."""
        self._meta = {}
        self._structure_version = 0
        self._graph = rx.PyDiGraph() if directed else rx.PyGraph()
        self._topology_mutable = True
        self._node_attr_mutable = True
        
    @property
    def nodes(self):
        """Return a view of the nodes that can be subscripted."""
        return GraphNodeView(self)
    
    @property
    def edges(self):
        """Return a view of the edges."""
        return GraphEdgeView(self)
    
    def __getitem__(self, node):
        """Get node data for a given node."""
        if not self._graph.has_node(node):
            raise KeyError(f"Node {node} not found in graph")
        node_data = self._graph.get_node_data(node)
        
        if not isinstance(node_data, dict):
            return MappingProxyType({})
        
        return MappingProxyType(node_data)

    def _is_topology_mutable(self):
        return getattr(self, '_topology_mutable', True)

    def _is_node_attr_mutable(self):
        return getattr(self, '_node_attr_mutable', True)

    def _set_mutability_policy(self, topology_mutable=True, node_attr_mutable=True):
        self._topology_mutable = bool(topology_mutable)
        self._node_attr_mutable = bool(node_attr_mutable)

    def _ensure_topology_mutable(self, op=None):
        if not self._is_topology_mutable():
            op_name = op or "topology mutation"
            raise PermissionError(f"{self.__class__.__name__} does not allow {op_name}")

    def _ensure_node_attr_mutable(self, op=None):
        if not self._is_node_attr_mutable():
            op_name = op or "node-attribute mutation"
            raise PermissionError(f"{self.__class__.__name__} does not allow {op_name}")

    def _normalize_node_attrs(self, node, attrs, op=None):
        return dict(attrs) if isinstance(attrs, dict) else {}

    def _validate_node_attrs(self, node, attrs, op=None):
        pass

    def _resolve_data_update_scope(self, node, changed_keys, op=None):
        return node

    def _before_node_data_mutation(self, node, attrs, scope_node=None, op=None):
        pass

    def _after_node_data_mutation(self, node, attrs, scope_node=None, op=None):
        pass

    def _apply_node_data_mutation(self, node, attrs, op=None, replace=False):
        if not self._graph.has_node(node):
            raise KeyError(f"Node {node} not found in graph")

        self._ensure_node_attr_mutable(op=op)
        normalized = self._normalize_node_attrs(node=node, attrs=attrs, op=op)
        self._validate_node_attrs(node=node, attrs=normalized, op=op)
        changed_keys = tuple(sorted(normalized.keys()))
        scope_node = self._resolve_data_update_scope(node=node, changed_keys=changed_keys, op=op)
        self._before_node_data_mutation(node=node, attrs=normalized, scope_node=scope_node, op=op)

        existing_data = self._graph.get_node_data(node)
        existing_data = existing_data if isinstance(existing_data, dict) else {}
        if replace:
            new_data = dict(normalized)
        else:
            new_data = dict(existing_data)
            new_data.update(normalized)
        self._graph[node] = new_data
        self._invalidate_caches()
        self._after_node_data_mutation(node=node, attrs=normalized, scope_node=scope_node, op=op)
    
    def __len__(self):
        """Return the number of nodes."""
        return self._graph.num_nodes()
    
    def __str__(self):
        """String representation of the graph."""
        return f"Graph with {self.number_of_nodes()} nodes and {self.number_of_edges()} edges"
    
    def __repr__(self):
        """String representation of the graph."""
        return f"Graph({self.number_of_nodes()}, {self.number_of_edges()})"
    
    def __iter__(self):
        """Iterate over node objects."""
        return iter(self._graph.node_indices())
    
    def __contains__(self, node):
        """Check if a node is in the graph."""
        return self._graph.has_node(node)
    
    def _invalidate_caches(self):
        """Invalidate all caches when structure changes"""
        self._structure_version += 1
        if hasattr(self, 'descendants'):
            self.descendants.cache_clear()
        if hasattr(self, 'ancestors'):
            self.ancestors.cache_clear()
        if hasattr(self, 'successors'):
            self.successors.cache_clear()
        if hasattr(self, 'predecessors'):
            self.predecessors.cache_clear()
    
    def out_degree(self, node):
        """Get the out-degree of a node"""
        if hasattr(self._graph, 'out_degree'):
            return self._graph.out_degree(node)
        else:
            return self._graph.degree(node)
    
    def in_degree(self, node):
        """Get the in-degree of a node"""
        if hasattr(self._graph, 'in_degree'):
            return self._graph.in_degree(node)
        else:
            return self._graph.degree(node)
    
    def _get_node_object(self, index):
        """Convert RustworkX node index to node object.
        
        For the base Graph class, nodes are just their indices.
        Subclasses can override this for different node representations.
        """
        return index
    
    def _get_node_index(self, node):
        """Convert node object to RustworkX index.
        
        For the base Graph class, nodes are just their indices.
        Subclasses can override this for different node representations.
        """
        return node

    @classmethod
    def directed(cls):
        """Create an empty directed graph."""
        return cls(directed=True)

    @classmethod
    def digraph(cls):
        """Alias for directed graph creation."""
        return cls.directed()

    @classmethod
    def from_rustworkx(cls, graph: Union[rx.PyGraph, rx.PyDiGraph], copy_graph: bool = True):
        """Create a Graph from an existing RustworkX graph."""
        if not isinstance(graph, (rx.PyGraph, rx.PyDiGraph)):
            raise TypeError(f"Expected rustworkx graph, got: {type(graph)}")

        new_graph = cls.__new__(cls)
        new_graph._graph = graph.copy() if copy_graph else graph
        new_graph._meta = {}
        new_graph._structure_version = 0
        new_graph._topology_mutable = True
        new_graph._node_attr_mutable = True
        return new_graph

    @classmethod
    def from_networkx(cls, graph, copy_graph: bool = True):
        """Create a Graph from a NetworkX graph."""
        import networkx as nx

        if not isinstance(graph, (nx.Graph, nx.DiGraph)):
            raise TypeError(f"Expected networkx.Graph or networkx.DiGraph, got: {type(graph)}")

        nx_graph = graph.copy() if copy_graph else graph
        directed = nx_graph.is_directed()
        new_graph = cls(directed=directed)

        node_index_map = {}
        for node, attrs in nx_graph.nodes(data=True):
            node_attrs = dict(attrs) if isinstance(attrs, dict) else {}
            node_index_map[node] = new_graph._graph.add_node(node_attrs)

        for u, v, attrs in nx_graph.edges(data=True):
            edge_attrs = dict(attrs) if isinstance(attrs, dict) else {}
            new_graph._graph.add_edge(node_index_map[u], node_index_map[v], edge_attrs)

        return new_graph

    @classmethod
    def from_nodes_edges(
        cls,
        nodes: Optional[List[Any]] = None,
        edges: Optional[List[Tuple[Any, ...]]] = None,
        directed: bool = False,
        node_mode: str = 'label',
        node_key: str = 'label',
    ):
        """Create a graph from node and edge iterables."""
        graph = cls(directed=directed)
        node_lookup: Dict[Any, int] = {}

        if node_mode not in {'label', 'id'}:
            raise ValueError("node_mode must be 'label' or 'id'")

        def ensure_node(node_value, attrs=None):
            attrs = dict(attrs) if isinstance(attrs, dict) else {}
            if node_mode == 'id':
                if not isinstance(node_value, int):
                    raise TypeError("node_mode='id' requires integer node values")
                while graph.number_of_nodes() <= node_value:
                    graph.add_node()
                graph.set_node_data(node_value, **attrs)
                return node_value

            if node_value in node_lookup:
                node_id = node_lookup[node_value]
                if attrs:
                    graph.set_node_data(node_id, **attrs)
                return node_id

            node_attrs = {node_key: node_value}
            node_attrs.update(attrs)
            node_id = graph.add_node(**node_attrs)
            node_lookup[node_value] = node_id
            return node_id

        if nodes:
            for node_entry in nodes:
                if isinstance(node_entry, tuple) and len(node_entry) == 2 and isinstance(node_entry[1], dict):
                    ensure_node(node_entry[0], node_entry[1])
                else:
                    ensure_node(node_entry)

        if edges:
            for edge in edges:
                if len(edge) == 2:
                    u_label, v_label = edge
                    edge_attrs = {}
                elif len(edge) == 3:
                    u_label, v_label, edge_attrs = edge
                    edge_attrs = dict(edge_attrs) if isinstance(edge_attrs, dict) else {}
                else:
                    raise ValueError("Edges must be (u, v) or (u, v, attrs)")

                u = ensure_node(u_label)
                v = ensure_node(v_label)
                graph.add_edge(u, v, **edge_attrs)

        return graph

    @classmethod
    def from_edges(
        cls,
        edges: List[Tuple[Any, ...]],
        directed: bool = False,
        node_mode: str = 'label',
        node_key: str = 'label',
    ):
        """Create a graph from an edge list."""
        return cls.from_nodes_edges(
            nodes=None,
            edges=edges,
            directed=directed,
            node_mode=node_mode,
            node_key=node_key,
        )

    @classmethod
    def empty_graph(
        cls,
        n_nodes: int = 0,
        labels: Optional[List[Any]] = None,
        directed: bool = False,
        node_key: str = 'label',
    ):
        """Create an empty graph with optional labeled nodes."""
        if n_nodes < 0:
            raise ValueError("n_nodes must be >= 0")
        resolved_labels = cls._resolve_labels(n_nodes, labels)
        return cls.from_nodes_edges(
            nodes=[(label, {node_key: label}) for label in resolved_labels],
            edges=[],
            directed=directed,
            node_mode='label',
            node_key=node_key,
        )

    @classmethod
    def path_graph(
        cls,
        n_nodes: int,
        labels: Optional[List[Any]] = None,
        directed: bool = False,
        node_key: str = 'label',
    ):
        """Create a path graph with optional labels."""
        if n_nodes < 0:
            raise ValueError("n_nodes must be >= 0")
        resolved_labels = cls._resolve_labels(n_nodes, labels)
        edges = [(resolved_labels[i], resolved_labels[i + 1]) for i in range(max(0, n_nodes - 1))]
        return cls.from_nodes_edges(
            nodes=[(label, {node_key: label}) for label in resolved_labels],
            edges=edges,
            directed=directed,
            node_mode='label',
            node_key=node_key,
        )

    @classmethod
    def cycle_graph(
        cls,
        n_nodes: int,
        labels: Optional[List[Any]] = None,
        directed: bool = False,
        node_key: str = 'label',
    ):
        """Create a cycle graph with optional labels."""
        if n_nodes < 0:
            raise ValueError("n_nodes must be >= 0")
        resolved_labels = cls._resolve_labels(n_nodes, labels)
        edges = [(resolved_labels[i], resolved_labels[i + 1]) for i in range(max(0, n_nodes - 1))]
        if n_nodes > 2:
            edges.append((resolved_labels[-1], resolved_labels[0]))
        return cls.from_nodes_edges(
            nodes=[(label, {node_key: label}) for label in resolved_labels],
            edges=edges,
            directed=directed,
            node_mode='label',
            node_key=node_key,
        )

    @classmethod
    def star_graph(
        cls,
        n_nodes: int,
        center: int = 0,
        labels: Optional[List[Any]] = None,
        directed: bool = False,
        node_key: str = 'label',
    ):
        """Create a star graph with optional labels."""
        if n_nodes < 0:
            raise ValueError("n_nodes must be >= 0")
        if n_nodes == 0:
            return cls.empty_graph(0, directed=directed, node_key=node_key)
        if center < 0 or center >= n_nodes:
            raise ValueError("center must be a valid node index")
        resolved_labels = cls._resolve_labels(n_nodes, labels)
        center_label = resolved_labels[center]
        edges = [(center_label, resolved_labels[i]) for i in range(n_nodes) if i != center]
        return cls.from_nodes_edges(
            nodes=[(label, {node_key: label}) for label in resolved_labels],
            edges=edges,
            directed=directed,
            node_mode='label',
            node_key=node_key,
        )

    @classmethod
    def random_graph(
        cls,
        n_nodes: int,
        p: float = 0.3,
        labels: Optional[List[Any]] = None,
        directed: bool = False,
        seed: Optional[int] = None,
        node_key: str = 'label',
    ):
        """Create a random Erdos-Renyi style graph with optional labels."""
        if n_nodes < 0:
            raise ValueError("n_nodes must be >= 0")
        if p < 0 or p > 1:
            raise ValueError("p must be between 0 and 1")
        resolved_labels = cls._resolve_labels(n_nodes, labels)
        rng = np.random.default_rng(seed)
        edges: List[Tuple[Any, Any]] = []
        if directed:
            for i in range(n_nodes):
                for j in range(n_nodes):
                    if i == j:
                        continue
                    if rng.random() < p:
                        edges.append((resolved_labels[i], resolved_labels[j]))
        else:
            for i in range(n_nodes):
                for j in range(i + 1, n_nodes):
                    if rng.random() < p:
                        edges.append((resolved_labels[i], resolved_labels[j]))
        return cls.from_nodes_edges(
            nodes=[(label, {node_key: label}) for label in resolved_labels],
            edges=edges,
            directed=directed,
            node_mode='label',
            node_key=node_key,
        )

    @staticmethod
    def _resolve_labels(n_nodes: int, labels: Optional[List[Any]] = None) -> List[Any]:
        if labels is None:
            return list(range(n_nodes))
        if len(labels) != n_nodes:
            raise ValueError(f"Expected {n_nodes} labels, got {len(labels)}")
        return list(labels)
    
    def neighbors(self, node):
        """Get neighbors of a node"""
        return list(self._graph.neighbors(node))
    
    @lru_cache(maxsize=None)
    def predecessors(self, node):
        """Returns all predecessors of a node.
        
        Parameters
        ----------
        node : int
            The node whose predecessors to return.
            
        Returns
        -------
        tuple
            All predecessors of the node.
        """
        _ = self._structure_version
        if hasattr(self._graph, 'predecessor_indices'):
            return tuple(self._graph.predecessor_indices(node))
        else:
            return tuple(self.neighbors(node))
    
    @lru_cache(maxsize=None)
    def successors(self, node):
        """Returns all successors of a node.
        
        Parameters
        ----------
        node : int
            The node whose successors to return.
            
        Returns
        -------
        tuple
            All successors of the node in sorted order (left-to-right).
        """
        _ = self._structure_version
        if hasattr(self._graph, 'successor_indices'):
            succ_indices = self._graph.successor_indices(node)
            return tuple(sorted(succ_indices))
        else:
            return tuple(sorted(self.neighbors(node)))
    
    @lru_cache(maxsize=None)
    def descendants(self, node):
        """Returns all descendants of a node using native RustworkX algorithm.
        
        Parameters
        ----------
        node : int
            The node whose descendants to return.
            
        Returns
        -------
        tuple
            All descendants of the node.
        """
        _ = self._structure_version
        try:
            return tuple(rx.descendants(self._graph, node))
        except Exception:
            return tuple()
    
    @lru_cache(maxsize=None)
    def ancestors(self, node):
        """Returns all ancestors of a node using native RustworkX algorithm.
        
        Parameters
        ----------
        node : int
            The node whose ancestors to return.
            
        Returns
        -------
        tuple
            All ancestors of the node.
        """
        _ = self._structure_version
        try:
            return tuple(rx.ancestors(self._graph, node))
        except Exception:
            return tuple()
    
    def topological_sort(self):
        """Returns nodes in topological order.
        
        Returns
        -------
        generator
            Nodes in topological order.
        """
        if hasattr(self._graph, 'out_degree'):
            indices = rx.topological_sort(self._graph)
        else:
            indices = self._graph.node_indices()
        
        return (idx for idx in indices)
    
    def to_directed(self):
        """Return a directed version of this graph.
        
        Returns
        -------
        Graph
            A new Graph instance with directed edges.
        """
        directed_rx = rx.PyDiGraph()
        
        for idx in self._graph.node_indices():
            node_data = self._graph.get_node_data(idx)
            directed_rx.add_node(node_data)
        
        for src, tgt, edge_data in self.edges(data=True):
            directed_rx.add_edge(src, tgt, edge_data)
        
        new_graph = Graph.__new__(Graph)
        new_graph._graph = directed_rx
        new_graph._meta = copy.deepcopy(self._meta)
        new_graph._structure_version = 0
        
        return new_graph
    
    def number_of_nodes(self):
        """Return the number of nodes in the graph.
        
        Returns
        -------
        int
            Number of nodes.
        """
        return self._graph.num_nodes()
        
    def number_of_edges(self):
        """Return the number of edges in the graph.
        
        Returns
        -------
        int
            Number of edges.
        """
        return self._graph.num_edges()
        
    def nodes_with_data(self, data=True):
        """Return nodes with their data.
        
        Parameters
        ----------
        data : bool
            If True, return node data as well.
            
        Returns
        -------
        Iterator
            Iterator of (node, data) pairs if data=True, else just nodes.
        """
        if data:
            for idx in self._graph.node_indices():
                node_data = self._graph.get_node_data(idx)
                yield (idx, node_data if isinstance(node_data, dict) else {})
        else:
            for idx in self._graph.node_indices():
                yield idx
    
    def subgraph(self, node, renumber=True):
        """Extract a subgraph starting from a given node.
        
        Parameters
        ----------
        node : int
            The node to use as the starting point of the subgraph.
        renumber : bool
            Whether to renumber the nodes in the new graph.
            
        Returns
        -------
        Graph
            A new Graph object representing the subgraph.
        """
        if node not in self:
            raise ValueError(f"Node {node} not found in graph")
            
        descendants = [node] + list(self.descendants(node))
        
        subgraph_rx = self._graph.subgraph(descendants)
        
        return self._from_graph(subgraph_rx, renumber=renumber)
    
    @property
    def root_nodes(self):
        """Returns root nodes (nodes with no predecessors)"""
        root_indices = []
        
        if hasattr(self._graph, 'in_degree'):
            for idx in self._graph.node_indices():
                if self._graph.in_degree(idx) == 0:
                    root_indices.append(idx)
        else:
            if self._graph.num_nodes() == 0:
                return tuple()
            
            degrees = [(idx, self._graph.degree(idx)) for idx in self._graph.node_indices()]
            if degrees:
                min_deg_nodes = [idx for idx, deg in degrees if deg > 0]
                if min_deg_nodes:
                    root_indices = [min(min_deg_nodes)]
        
        return tuple(root_indices)
    
    def add_node(self, **attr):
        """Add a node to the graph.
        
        Parameters
        ----------
        **attr : dict
            Node attributes as keyword arguments.
            
        Returns
        -------
        int
            The node ID that was added.
        """
        self._ensure_topology_mutable(op='add_node')
        node_id = self._graph.add_node(attr if attr else {})
        self._invalidate_caches()
        return node_id
    
    def set_node_data(self, node, **attr):
        """Update data for an existing node.
        
        Parameters
        ----------
        node : int
            The node to update.
        **attr : dict
            Node attributes to set.
        """
        self._apply_node_data_mutation(node=node, attrs=attr, op='set_node_data', replace=False)

    def update_node_data(self, node, attrs: Dict[str, Any]):
        """Update data for an existing node from a dictionary."""
        self._apply_node_data_mutation(node=node, attrs=attrs, op='update_node_data', replace=False)

    def replace_node_data(self, node, attrs: Dict[str, Any]):
        """Replace all data for an existing node."""
        self._apply_node_data_mutation(node=node, attrs=attrs, op='replace_node_data', replace=True)
    
    def remove_node(self, node):
        """Remove a node from the graph."""
        self._ensure_topology_mutable(op='remove_node')
        self._graph.remove_node(node)
        self._invalidate_caches()
        
    def add_edge(self, u, v, **attr):
        """Add an edge to the graph with optional attributes."""
        self._ensure_topology_mutable(op='add_edge')
        if not self._graph.has_node(u):
            raise KeyError(f"Node {u} not found in graph")
        if not self._graph.has_node(v):
            raise KeyError(f"Node {v} not found in graph")
        
        self._graph.add_edge(u, v, attr if attr else {})
        self._invalidate_caches()
        
    def has_edge(self, u, v):
        """Check if an edge exists between two nodes."""
        return self._graph.has_edge(u, v)
        
    def remove_edge(self, u, v):
        """Remove an edge from the graph."""
        self._ensure_topology_mutable(op='remove_edge')
        self._graph.remove_edge(u, v)
        self._invalidate_caches()
        
    def update(self, edges=None, nodes=None):
        """Update the graph with nodes and edges."""
        self._ensure_topology_mutable(op='update')
        if nodes:
            for node_data in nodes:
                if isinstance(node_data, tuple) and len(node_data) == 2:
                    node, attrs = node_data
                    if not self._graph.has_node(node):
                        raise KeyError(f"Node {node} not found in graph")
                    self.set_node_data(node, **attrs)
                else:
                    self.add_node(label=node_data)
        
        if edges:
            for edge_data in edges:
                if len(edge_data) == 2:
                    u, v = edge_data
                    self.add_edge(u, v)
                elif len(edge_data) == 3:
                    u, v, attrs = edge_data
                    self.add_edge(u, v, **attrs)
        
        self._invalidate_caches()
        
    def clear(self):
        """Remove all nodes and edges from the graph."""
        self._ensure_topology_mutable(op='clear')
        self._graph.clear()
        self._invalidate_caches()
    
    def set_node_attributes(self, node, attributes):
        """Set attributes for a node."""
        self._apply_node_data_mutation(node=node, attrs=attributes, op='set_node_attributes', replace=False)
    
    def clear_node_attributes(self, nodes=None):
        """Clear attributes of specified nodes or all nodes.
        
        Parameters
        ----------
        nodes : list, optional
            Specific nodes to clear attributes for, or None for all nodes.
        """
        self._ensure_node_attr_mutable(op='clear_node_attributes')
        nodes_to_clear = nodes if nodes is not None else self._graph.node_indices()
        for node in nodes_to_clear:
            if node in self:
                self._apply_node_data_mutation(node=node, attrs={}, op='clear_node_attributes', replace=True)
        
    def renumber_nodes(self, method='default'):
        """Renumber the nodes in the graph to consecutive integers.
        
        Parameters
        ----------
        method : str
            The method to use for renumbering:
            - 'default': Use sequential numbering
            - 'dfs': Use depth-first search preorder
            - 'bfs': Use breadth-first search
                
        Returns
        -------
        Graph
            Self with renumbered nodes.
        """
        if method == 'default':
            pass
        elif method in ['dfs', 'bfs']:
            pass
        else:
            raise ValueError(f"Unknown renumbering method: {method}")
            
        return self

    def copy(self):
        """Create a deep copy of this graph."""
        return copy.deepcopy(self)
    
    def is_directed(self):
        """Return True if graph is directed, False otherwise."""
        return isinstance(self._graph, rx.PyDiGraph)
    
    def is_multigraph(self):
        """Return True if graph is a multigraph, False otherwise."""
        return False
    
    def to_networkx(self):
        """Convert this Graph to a NetworkX graph for compatibility with NetworkX functions.
        
        Returns
        -------
        networkx.Graph or networkx.DiGraph
            NetworkX equivalent of this graph.
        """
        import networkx as nx
        
        if self.is_directed():
            nx_graph = nx.DiGraph()
        else:
            nx_graph = nx.Graph()
        
        for node, attrs in self.nodes(data=True):
            nx_graph.add_node(node, **attrs)
        
        for u, v, attrs in self.edges(data=True):
            nx_graph.add_edge(u, v, **attrs)
        
        return nx_graph

    @classmethod
    def _from_graph(cls, G, **kwargs):
        """Create a new instance from an existing graph.
        
        Parameters
        ----------
        G : Graph or rustworkx graph
            The graph to create a new instance from.
        **kwargs : dict
            Additional arguments.
            
        Returns
        -------
        Graph
            A new Graph instance.
        """
        if isinstance(G, Graph):
            new_graph = cls.from_rustworkx(G._graph)
            new_graph._meta = copy.deepcopy(G._meta)
            return new_graph
        if isinstance(G, (rx.PyGraph, rx.PyDiGraph)):
            return cls.from_rustworkx(G)
        raise TypeError(f"Unsupported graph type: {type(G)}")
    
    @classmethod
    def grid_graph(cls, dims, periodic=False):
        """Create a grid graph with dimensions specified in dims.
        
        Creates a grid graph where nodes are numbered sequentially 0, 1, 2, ...
        and node data contains the coordinate as a tuple.
        
        Parameters
        ----------
        dims : list or tuple
            List of ranges or iterables defining the coordinate space for each dimension.
        periodic : bool, optional
            Whether to create periodic boundary conditions (default is False)
            
        Returns
        -------
        Graph
            A new Graph instance with grid structure and coordinate data in nodes
        """
        if not dims or len(dims) == 0:
            return cls()
        
        dims = [list(d) for d in dims]
        
        if len(dims) == 1:
            if periodic:
                rx_graph = rx.generators.cycle_graph(len(dims[0]))
            else:
                rx_graph = rx.generators.path_graph(len(dims[0]))
            for i, coord_val in enumerate(dims[0]):
                rx_graph[i] = {'coord': (coord_val,)}
        elif len(dims) == 2 and not periodic:
            rows, cols = len(dims[0]), len(dims[1])
            rx_graph = rx.generators.grid_graph(rows, cols)
            import itertools
            for i, coord in enumerate(itertools.product(dims[0], dims[1])):
                rx_graph[i] = {'coord': coord}
        else:
            factor_graphs = []
            for dim_values in dims:
                if periodic:
                    factor_graph = rx.generators.cycle_graph(len(dim_values))
                else:
                    factor_graph = rx.generators.path_graph(len(dim_values))
                factor_graphs.append(factor_graph)

            product_graph = factor_graphs[0]
            id_to_coord_index = {node_id: (node_id,) for node_id in product_graph.node_indices()}

            for next_factor in factor_graphs[1:]:
                product_graph, node_map = rx.graph_cartesian_product(product_graph, next_factor)
                next_id_to_coord_index = {}
                for (left_node, right_node), product_node in node_map.items():
                    next_id_to_coord_index[product_node] = id_to_coord_index[left_node] + (right_node,)
                id_to_coord_index = next_id_to_coord_index

            for node_id, coord_idx_tuple in id_to_coord_index.items():
                coord = tuple(dims[i][coord_idx_tuple[i]] for i in range(len(dims)))
                product_graph[node_id] = {'coord': coord}

            rx_graph = product_graph
        
        return cls.from_rustworkx(rx_graph)
    
    @classmethod
    def complete_graph(cls, n_nodes, labels: Optional[List[Any]] = None, directed: bool = False, node_key: str = 'label'):
        """Create a complete graph.
        
        Parameters
        ----------
        n_nodes : int
            Number of nodes in the complete graph.
            
        Returns
        -------
        Graph
            A new Graph instance with complete structure.
        """
        if n_nodes < 0:
            raise ValueError("n_nodes must be >= 0")
        resolved_labels = cls._resolve_labels(n_nodes, labels)
        edges: List[Tuple[Any, Any]] = []
        if directed:
            for i in range(n_nodes):
                for j in range(n_nodes):
                    if i != j:
                        edges.append((resolved_labels[i], resolved_labels[j]))
        else:
            for i in range(n_nodes):
                for j in range(i + 1, n_nodes):
                    edges.append((resolved_labels[i], resolved_labels[j]))
        return cls.from_nodes_edges(
            nodes=[(label, {node_key: label}) for label in resolved_labels],
            edges=edges,
            directed=directed,
            node_mode='label',
            node_key=node_key,
        )
    
    @classmethod
    def from_cost_matrix(cls, cost_matrix: np.ndarray, items: List[T]):
        """Create a Graph from a cost matrix.
        
        Transform a symmetric cost matrix into an undirected graph where
        edge weights represent the costs between nodes. Self-loops are
        excluded from the resulting graph.

        Parameters
        ----------
        cost_matrix : numpy.ndarray
            Symmetric cost matrix with numeric values. Should be square
            with dimensions matching the length of items.
        items : List[T]
            List of items corresponding to matrix indices. Used as node
            values in the resulting graph.

        Returns
        -------
        Graph
            Undirected graph with nodes corresponding to matrix indices
            and edge weights equal to the cost matrix values. Only edges
            with positive costs are included. Node attributes 'value' 
            contain the original items.

        Examples
        --------
        Create a graph from a simple cost matrix:
        
        >>> import numpy as np
        >>> matrix = np.array([[0, 1, 2], [1, 0, 3], [2, 3, 0]])
        >>> items = ['A', 'B', 'C']
        >>> graph = Graph.from_cost_matrix(matrix, items)
        >>> list(graph.edges(data=True))
        [('A', 'B', {'weight': 1}), ('A', 'C', {'weight': 2}), ('B', 'C', {'weight': 3})]
        """
        graph = cls()
        
        node_list = []
        for item in items:
            node_id = graph.add_node(value=item)
            node_list.append((node_id, {'value': item}))
        
        edge_list = []
        for i in range(len(items)):
            for j in range(i + 1, len(items)):
                cost = cost_matrix[i, j]
                if cost > 0:
                    edge_list.append((node_list[i][0], node_list[j][0], {'weight': cost}))
        
        if edge_list:
            graph._graph.add_edges_from(edge_list)
        
        return graph
    
    def __deepcopy__(self, memo):
        new_graph = self.__class__.__new__(self.__class__)
        
        new_graph._graph = self._graph.copy()
        new_graph._meta = copy.deepcopy(self._meta, memo)
        new_graph._structure_version = 0
        new_graph._topology_mutable = self._is_topology_mutable()
        new_graph._node_attr_mutable = self._is_node_attr_mutable()
        
        return new_graph


class GraphNodeView:
    """View of graph nodes that mimics NetworkX NodeView behavior."""
    
    def __init__(self, graph: Graph):
        self._graph = graph
    
    def __iter__(self):
        return iter(self._graph._graph.node_indices())
    
    def __len__(self):
        return self._graph.number_of_nodes()
    
    def __contains__(self, node):
        return self._graph._graph.has_node(node)
    
    def __getitem__(self, node):
        node_data = self._graph._graph.get_node_data(node)
        if not isinstance(node_data, dict):
            return MappingProxyType({})
        return MappingProxyType(node_data)
    
    def __call__(self, data=False):
        """Return nodes with optional data."""
        if data:
            for idx in self._graph._graph.node_indices():
                node_data = self._graph._graph.get_node_data(idx)
                if isinstance(node_data, dict):
                    yield (idx, MappingProxyType(node_data))
                else:
                    yield (idx, MappingProxyType({}))
        else:
            for idx in self._graph._graph.node_indices():
                yield idx


class GraphEdgeView:
    """View of graph edges that mimics NetworkX EdgeView behavior."""
    
    def __init__(self, graph: Graph):
        self._graph = graph
    
    def __iter__(self):
        for edge_data in self._graph._graph.edge_list():
            src_idx, tgt_idx = edge_data
            yield (src_idx, tgt_idx)
    
    def __len__(self):
        return self._graph.number_of_edges()
    
    def __call__(self, data=False):
        """Return edges with optional data."""
        if data:
            for src_idx, tgt_idx in self._graph._graph.edge_list():
                edge_data = self._graph._graph.get_edge_data(src_idx, tgt_idx)
                yield (src_idx, tgt_idx, edge_data if isinstance(edge_data, dict) else {})
        else:
            for src_idx, tgt_idx in self._graph._graph.edge_list():
                yield (src_idx, tgt_idx)
    
    def __getitem__(self, edge):
        """Get edge data for a given edge (u, v)."""
        u, v = edge
        edge_data = self._graph._graph.get_edge_data(u, v)
        return edge_data if isinstance(edge_data, dict) else {}
    


import math
from klotho.dynatos.dynamics import freq_amp_scale, ampdb


DEFAULT_AMP = 0.5


def single_voice_amplitude(freq, target_amp=None):
    if target_amp is None:
        target_amp = DEFAULT_AMP
    scaled = freq_amp_scale(freq, ampdb(target_amp))
    return min(1.0, max(0.05, target_amp * (scaled / max(0.01, target_amp))))


def compute_voice_amplitudes(freqs, target_amp=None):
    if target_amp is None:
        target_amp = DEFAULT_AMP
    n = len(freqs)
    if n == 0:
        return []
    if n == 1:
        return [single_voice_amplitude(freqs[0], target_amp)]

    weights = []
    for freq in freqs:
        w = freq_amp_scale(freq, ampdb(target_amp))
        weights.append(max(0.01, w))

    total_w = sum(weights)
    normalized = [w / total_w for w in weights]

    headroom = max(1.0, math.sqrt(n))
    amps = [(target_amp * nw * n) / headroom for nw in normalized]

    return [min(1.0, max(0.05, a)) for a in amps]



import copy

from klotho.thetos.instruments.presets import TONEJS_PRESETS
from klotho.utils.data_structures.dictionaries import SafeDict

from .base import Instrument


class ToneInstrument(Instrument):
    def __init__(self, name='Synth', tonejs_class='Synth', pfields=None):
        if pfields is None:
            pfields = {}
        base_pfields = dict(pfields)
        if 'freq' not in base_pfields:
            base_pfields['freq'] = 440.0
        if 'vel' in base_pfields:
            vel = base_pfields['vel']
        elif 'amp' in base_pfields:
            vel = base_pfields['amp']
        else:
            vel = 0.6
        base_pfields['vel'] = vel
        base_pfields.pop('amp', None)
        safe_pfields = SafeDict(base_pfields, aliases={'amp': 'vel'})
        super().__init__(name=name, pfields=safe_pfields)
        self._tonejs_class = tonejs_class

    @property
    def tonejs_class(self):
        return self._tonejs_class

    def __str__(self):
        return f"ToneInstrument(name='{self._name}', tonejs_class='{self._tonejs_class}', pfields={dict(self._pfields)})"

    @classmethod
    def _validate_kwargs(cls, valid_keys, kwargs):
        allowed = set(valid_keys) | {'freq', 'vel', 'amp'}
        invalid = set(kwargs.keys()) - allowed
        if invalid:
            raise ValueError(
                f"Invalid pfield(s): {invalid}. "
                f"Valid pfields: {sorted(allowed)}"
            )

    @classmethod
    def from_preset(cls, preset_name: str, name: str = None, **kwargs):
        preset = TONEJS_PRESETS[preset_name]
        pfields = copy.deepcopy(preset['pfields'])
        if kwargs:
            cls._validate_kwargs(pfields.keys(), kwargs)
            pfields.update(kwargs)
        return cls(name=name or preset_name, tonejs_class=preset['tonejs_class'], pfields=pfields)

    @classmethod
    def Harmonics(cls, name=None, **kwargs):
        return cls.from_preset('Harmonics', name=name, **kwargs)

    @classmethod
    def Tiny(cls, name=None, **kwargs):
        return cls.from_preset('Tiny', name=name, **kwargs)

    @classmethod
    def ElectricCello(cls, name=None, **kwargs):
        return cls.from_preset('ElectricCello', name=name, **kwargs)

    @classmethod
    def Kalimba(cls, name=None, **kwargs):
        return cls.from_preset('Kalimba', name=name, **kwargs)

    @classmethod
    def ThinSaws(cls, name=None, **kwargs):
        return cls.from_preset('ThinSaws', name=name, **kwargs)

    @classmethod
    def Bah(cls, name=None, **kwargs):
        return cls.from_preset('Bah', name=name, **kwargs)

    @classmethod
    def BassGuitar(cls, name=None, **kwargs):
        return cls.from_preset('BassGuitar', name=name, **kwargs)

    @classmethod
    def Bassy(cls, name=None, **kwargs):
        return cls.from_preset('Bassy', name=name, **kwargs)

    @classmethod
    def BrassCircuit(cls, name=None, **kwargs):
        return cls.from_preset('BrassCircuit', name=name, **kwargs)

    @classmethod
    def CoolGuy(cls, name=None, **kwargs):
        return cls.from_preset('CoolGuy', name=name, **kwargs)

    @classmethod
    def Pianoetta(cls, name=None, **kwargs):
        return cls.from_preset('Pianoetta', name=name, **kwargs)

    @classmethod
    def Pizz(cls, name=None, **kwargs):
        return cls.from_preset('Pizz', name=name, **kwargs)

    @classmethod
    def Gravel(cls, name=None, **kwargs):
        return cls.from_preset('Gravel', name=name, **kwargs)

    @classmethod
    def Slap(cls, name=None, **kwargs):
        return cls.from_preset('Slap', name=name, **kwargs)

    @classmethod
    def Swoosh(cls, name=None, **kwargs):
        return cls.from_preset('Swoosh', name=name, **kwargs)

    @classmethod
    def Train(cls, name=None, **kwargs):
        return cls.from_preset('Train', name=name, **kwargs)

    @classmethod
    def AlienChorus(cls, name=None, **kwargs):
        return cls.from_preset('AlienChorus', name=name, **kwargs)

    @classmethod
    def DelicateWindPart(cls, name=None, **kwargs):
        return cls.from_preset('DelicateWindPart', name=name, **kwargs)

    @classmethod
    def DropPulse(cls, name=None, **kwargs):
        return cls.from_preset('DropPulse', name=name, **kwargs)

    @classmethod
    def Lectric(cls, name=None, **kwargs):
        return cls.from_preset('Lectric', name=name, **kwargs)

    @classmethod
    def Marimba(cls, name=None, **kwargs):
        return cls.from_preset('Marimba', name=name, **kwargs)

    @classmethod
    def Steelpan(cls, name=None, **kwargs):
        return cls.from_preset('Steelpan', name=name, **kwargs)

    @classmethod
    def SuperSaw(cls, name=None, **kwargs):
        return cls.from_preset('SuperSaw', name=name, **kwargs)

    @classmethod
    def TreeTrunk(cls, name=None, **kwargs):
        return cls.from_preset('TreeTrunk', name=name, **kwargs)

    @classmethod
    def Kick(cls, name=None, **kwargs):
        pfields = {
            'freq': 52, 'vel': 0.9, 'tuneHz': 52,
            'decay': 0.35, 'pitchDecay': 0.02, 'punch': 6, 'click': 0.25
        }
        if kwargs:
            cls._validate_kwargs(pfields.keys(), kwargs)
            pfields.update(kwargs)
        return cls(name=name or 'Kick', tonejs_class='Kick', pfields=pfields)

    @classmethod
    def Snare(cls, name=None, **kwargs):
        pfields = {
            'freq': 190, 'vel': 0.85, 'tuneHz': 190,
            'decay': 0.18, 'snap': 0.9, 'body': 0.45, 'toneHz': 1800
        }
        if kwargs:
            cls._validate_kwargs(pfields.keys(), kwargs)
            pfields.update(kwargs)
        return cls(name=name or 'Snare', tonejs_class='Snare', pfields=pfields)

    @classmethod
    def TomLow(cls, name=None, **kwargs):
        pfields = {
            'freq': 110, 'vel': 0.75, 'tuneHz': 110,
            'decay': 0.35, 'pitchDecay': 0.01, 'punch': 4
        }
        if kwargs:
            cls._validate_kwargs(pfields.keys(), kwargs)
            pfields.update(kwargs)
        return cls(name=name or 'TomLow', tonejs_class='TomLow', pfields=pfields)

    @classmethod
    def TomMid(cls, name=None, **kwargs):
        pfields = {
            'freq': 160, 'vel': 0.75, 'tuneHz': 160,
            'decay': 0.35, 'pitchDecay': 0.01, 'punch': 4
        }
        if kwargs:
            cls._validate_kwargs(pfields.keys(), kwargs)
            pfields.update(kwargs)
        return cls(name=name or 'TomMid', tonejs_class='TomMid', pfields=pfields)

    @classmethod
    def TomHigh(cls, name=None, **kwargs):
        pfields = {
            'freq': 220, 'vel': 0.75, 'tuneHz': 220,
            'decay': 0.35, 'pitchDecay': 0.01, 'punch': 4
        }
        if kwargs:
            cls._validate_kwargs(pfields.keys(), kwargs)
            pfields.update(kwargs)
        return cls(name=name or 'TomHigh', tonejs_class='TomHigh', pfields=pfields)

    @classmethod
    def HatClosed(cls, name=None, **kwargs):
        pfields = {
            'freq': 420, 'vel': 0.55, 'decay': 0.05,
            'resonance': 5200, 'harmonicity': 5.1, 'modulationIndex': 32,
            'frequency': 420, 'octaves': 1.5
        }
        if kwargs:
            cls._validate_kwargs(pfields.keys(), kwargs)
            pfields.update(kwargs)
        return cls(name=name or 'HatClosed', tonejs_class='HatClosed', pfields=pfields)

    @classmethod
    def HatOpen(cls, name=None, **kwargs):
        pfields = {
            'freq': 420, 'vel': 0.45, 'decay': 0.45,
            'resonance': 5200, 'harmonicity': 5.1, 'modulationIndex': 32,
            'frequency': 420, 'octaves': 1.5
        }
        if kwargs:
            cls._validate_kwargs(pfields.keys(), kwargs)
            pfields.update(kwargs)
        return cls(name=name or 'HatOpen', tonejs_class='HatOpen', pfields=pfields)

    @classmethod
    def Crash(cls, name=None, **kwargs):
        pfields = {
            'freq': 320, 'vel': 0.55, 'decay': 2.8,
            'resonance': 5200, 'harmonicity': 3.7, 'modulationIndex': 18,
            'frequency': 320, 'octaves': 2.2
        }
        if kwargs:
            cls._validate_kwargs(pfields.keys(), kwargs)
            pfields.update(kwargs)
        return cls(name=name or 'Crash', tonejs_class='Crash', pfields=pfields)

    @classmethod
    def Ride(cls, name=None, **kwargs):
        pfields = {
            'freq': 280, 'vel': 0.35, 'decay': 1.7,
            'resonance': 4500, 'harmonicity': 4.2, 'modulationIndex': 12,
            'frequency': 280, 'octaves': 2.0
        }
        if kwargs:
            cls._validate_kwargs(pfields.keys(), kwargs)
            pfields.update(kwargs)
        return cls(name=name or 'Ride', tonejs_class='Ride', pfields=pfields)



from typing import Union
from fractions import Fraction
from klotho.utils.data_structures.enums import MinMaxEnum

__all__ = [
    'TEMPO',
    'metric_modulation',
    'tempo_for_duration',
    'beat_for_duration',
]

class TEMPO(MinMaxEnum):
    """
    Enum for musical tempo markings mapped to beats per minute (bpm).

    Each tempo marking is associated with a range of beats per minute.
    This enumeration returns a tuple representing the minimum and maximum
    bpm for each tempo.

    .. list-table:: Tempo Markings
       :header-rows: 1

       * - Name
         - Tempo Marking
         - BPM Range
       * - Larghissimo
         - extremely slow
         - 12 -- 24
       * - Adagissimo_Grave
         - very slow, solemn
         - 24 -- 40
       * - Largo
         - slow and broad
         - 40 -- 66
       * - Larghetto
         - rather slow and broad
         - 44 -- 66
       * - Adagio
         - slow and expressive
         - 44 -- 68
       * - Adagietto
         - slower than andante
         - 46 -- 80
       * - Lento
         - slow
         - 52 -- 108
       * - Andante
         - walking pace
         - 56 -- 108
       * - Andantino
         - slightly faster than andante
         - 80 -- 108
       * - Marcia_Moderato
         - moderate march
         - 66 -- 80
       * - Andante_Moderato
         - between andante and moderato
         - 80 -- 108
       * - Moderato
         - moderate speed
         - 108 -- 120
       * - Allegretto
         - moderately fast
         - 112 -- 120
       * - Allegro_Moderato
         - slightly less than allegro
         - 116 -- 120
       * - Allegro
         - fast, bright
         - 120 -- 156
       * - Molto_Allegro_Allegro_Vivace
         - slightly faster than allegro
         - 124 -- 156
       * - Vivace
         - lively, fast
         - 156 -- 176
       * - Vivacissimo_Allegrissimo
         - very fast, bright
         - 172 -- 176
       * - Presto
         - very fast
         - 168 -- 200
       * - Prestissimo
         - extremely fast
         - 200 -- 300

    Examples
    --------
    >>> TEMPO.Adagio.min
    44
    >>> TEMPO.Adagio.max
    68
    """
    Larghissimo                  = (12, 24)
    Adagissimo_Grave             = (24, 40)
    Largo                        = (40, 66)
    Larghetto                    = (44, 66)
    Adagio                       = (44, 68)
    Adagietto                    = (46, 80)
    Lento                        = (52, 108)
    Andante                      = (56, 108)
    Andantino                    = (80, 108)
    Marcia_Moderato              = (66, 80)
    Andante_Moderato             = (80, 108)
    Moderato                     = (108, 120)
    Allegretto                   = (112, 120)
    Allegro_Moderato             = (116, 120)
    Allegro                      = (120, 156)
    Molto_Allegro_Allegro_Vivace = (124, 156)
    Vivace                       = (156, 176)
    Vivacissimo_Allegrissimo     = (172, 176)
    Presto                       = (168, 200)
    Prestissimo                  = (200, 300)
  
def metric_modulation(current_tempo:float, current_beat_value:Union[Fraction,str,float], new_beat_value:Union[Fraction,str,float]) -> float:
    """
    Determine the new tempo for a metric modulation between two beat values.

    Metric modulation maintains the duration of a beat constant while changing
    the note value that represents the beat, effectively changing the tempo.

    See: https://en.wikipedia.org/wiki/Metric_modulation

    Parameters
    ----------
    current_tempo : float
        The original tempo in beats per minute.
    current_beat_value : Fraction, str, or float
        The note value (as a fraction of a whole note) representing one beat
        before modulation.
    new_beat_value : Fraction, str, or float
        The note value (as a fraction of a whole note) representing one beat
        after modulation.

    Returns
    -------
    float
        The new tempo in beats per minute after the metric modulation.

    Examples
    --------
    >>> metric_modulation(120, '1/4', '1/8')
    240.0
    """
    current_beat_value = Fraction(current_beat_value)
    new_beat_value = Fraction(new_beat_value)
    current_duration = 60 / current_tempo * current_beat_value
    new_tempo = 60 / current_duration * new_beat_value
    return float(new_tempo)

def tempo_for_duration(metric_ratio: Union[Fraction, str, float], reference_beat: Union[Fraction, str, float], duration: float) -> float:
    """
    Calculate the tempo required for a metric ratio to last a specified duration.

    Parameters
    ----------
    metric_ratio : Fraction, str, or float
        The metric ratio representing the total duration (e.g., ``'4/4'``).
    reference_beat : Fraction, str, or float
        The beat value that defines the tempo (e.g., ``'1/4'`` for a
        quarter note).
    duration : float
        The desired duration in seconds.

    Returns
    -------
    float
        The tempo in beats per minute.

    Examples
    --------
    >>> tempo_for_duration('4/4', '1/4', 2.0)
    120.0
    """
    metric_ratio = Fraction(metric_ratio)
    reference_beat = Fraction(reference_beat)
    
    beats_in_metric = metric_ratio / reference_beat
    bpm = float(beats_in_metric * 60 / duration)
    
    return bpm

def beat_for_duration(metric_ratio: Union[Fraction, str, float], bpm: float, duration: float) -> Fraction:
    """
    Calculate the reference beat value for a metric ratio at a given tempo and duration.

    Parameters
    ----------
    metric_ratio : Fraction, str, or float
        The metric ratio representing the total duration (e.g., ``'4/4'``).
    bpm : float
        The tempo in beats per minute.
    duration : float
        The desired duration in seconds.

    Returns
    -------
    Fraction
        The reference beat value as a fraction.

    Examples
    --------
    >>> beat_for_duration('4/4', 120, 2.0)
    Fraction(1, 4)
    """
    metric_ratio = Fraction(metric_ratio)
    reference_beat = Fraction(metric_ratio * 60) / Fraction(bpm * duration)
    
    return reference_beat



import math
import numpy as np
import networkx as nx
from sklearn.manifold import MDS, SpectralEmbedding

from klotho.topos.graphs.lattices import Lattice

from .._shared.colors import SHAPE_COLORS as _SHAPE_COLORS_GLOBAL


def _coerce_to_tuples(items):
    if items is None:
        return None
    result = []
    for item in items:
        if isinstance(item, (list, tuple)):
            result.append(tuple(item))
        elif hasattr(item, 'tolist'):
            result.append(tuple(item.tolist()))
        elif hasattr(item, '__iter__') and not isinstance(item, str):
            result.append(tuple(item))
        else:
            result.append(item)
    return result


def _prepare_lattice_coordinates(lattice, nodes, path, path_mode, fit, pad,
                                 dim_reduction, target_dims,
                                 mds_metric, mds_max_iter,
                                 spectral_affinity, spectral_gamma):
    if lattice.dimensionality > 3 and dim_reduction is None:
        raise ValueError(
            f"Plotting dimensionality > 3 requires dim_reduction. "
            f"Got dimensionality={lattice.dimensionality}. "
            f"Use dim_reduction='mds' or 'spectral'"
        )
    if target_dims not in [2, 3]:
        raise ValueError(f"target_dims must be 2 or 3, got {target_dims}")

    if lattice.dimensionality <= 2:
        max_resolution = 5
    elif lattice.dimensionality == 3:
        max_resolution = 2
    elif target_dims == 3:
        max_resolution = 1
    else:
        max_resolution = 3

    expected_total = 1
    for dim in lattice._dims:
        expected_total *= len(dim)
        if expected_total > 10000:
            expected_total = float('inf')
            break

    has_selection = bool(nodes) or bool(path)

    fit_mode = None
    if fit is True:
        fit_mode = 'rect'
    elif isinstance(fit, str) and fit in ('rect', 'square', 'tight'):
        fit_mode = fit

    if has_selection and fit_mode:
        import itertools
        all_coords_to_fit = list(nodes or []) + list(path or [])

        if all_coords_to_fit:
            ndim = lattice.dimensionality
            p = max(0, int(pad))

            if fit_mode == 'tight':
                selection_set = set(all_coords_to_fit)
                neighbor_shell = set()
                for coord in selection_set:
                    for d in range(ndim):
                        for delta in range(-p, p + 1):
                            if delta == 0:
                                continue
                            neighbor = list(coord)
                            neighbor[d] += delta
                            neighbor = tuple(neighbor)
                            if neighbor in lattice:
                                neighbor_shell.add(neighbor)
                coords = sorted(selection_set | neighbor_shell)
            else:
                per_dim_lo = []
                per_dim_hi = []
                for d in range(ndim):
                    vals = [c[d] for c in all_coords_to_fit]
                    per_dim_lo.append(min(vals))
                    per_dim_hi.append(max(vals))

                if fit_mode == 'square':
                    half_spans = [(hi - lo) / 2.0 for lo, hi in zip(per_dim_lo, per_dim_hi)]
                    max_half = max(half_spans)
                    centers = [(lo + hi) / 2.0 for lo, hi in zip(per_dim_lo, per_dim_hi)]
                    for d in range(ndim):
                        per_dim_lo[d] = int(centers[d] - max_half)
                        per_dim_hi[d] = int(centers[d] + max_half) + (1 if (max_half % 1) else 0)

                ranges = []
                for d in range(ndim):
                    lo = max(per_dim_lo[d] - p, min(lattice._dims[d]))
                    hi = min(per_dim_hi[d] + p, max(lattice._dims[d]))
                    ranges.append(range(lo, hi + 1))

                coords = [c for c in itertools.product(*ranges) if c in lattice]
        else:
            coords = lattice.coords

    elif lattice.dimensionality > 3 or lattice._is_lazy or expected_total > 1000:
        if path:
            import itertools
            coord_ranges = []
            for dim_idx in range(lattice.dimensionality):
                dim_vals = [coord[dim_idx] for coord in path if coord in lattice]
                if dim_vals:
                    min_val, max_val = min(dim_vals), max(dim_vals)
                    coord_ranges.append((min_val - 2, max_val + 2))
                else:
                    coord_ranges.append((-max_resolution, max_resolution))
            ranges = [range(start, end + 1) for start, end in coord_ranges]
            coords = [coord for coord in itertools.product(*ranges) if coord in lattice]
        else:
            coords = lattice._get_plot_coords(max_resolution)
    else:
        coords = lattice.coords

    if has_selection or lattice.dimensionality > 3 or lattice._is_lazy or expected_total > 1000:
        G_reduced = nx.Graph()
        coord_set = set(coords)
        G_reduced.add_nodes_from(coords)
        ndim = lattice.dimensionality
        for coord in coords:
            for d in range(ndim):
                neighbor = list(coord)
                neighbor[d] += 1
                neighbor = tuple(neighbor)
                if neighbor in coord_set:
                    G_reduced.add_edge(coord, neighbor)
        G = G_reduced
    else:
        G = lattice

    if nodes and path_mode == 'origin':
        origin = tuple(0 for _ in range(lattice.dimensionality))
        if origin not in coords:
            coords.append(origin)
            if hasattr(G, 'add_node'):
                G.add_node(origin)
                for coord in coords:
                    if coord != origin:
                        diff_count = sum(1 for a, b in zip(origin, coord) if abs(a - b) == 1)
                        same_count = sum(1 for a, b in zip(origin, coord) if a == b)
                        if diff_count == 1 and same_count == len(coord) - 1:
                            G.add_edge(origin, coord)

    original_coords = coords

    if lattice.dimensionality > 3:
        coord_matrix = np.array([list(coord) for coord in coords])

        if dim_reduction == 'mds':
            reducer = MDS(n_components=target_dims, metric=mds_metric, max_iter=mds_max_iter, random_state=42)
            reduced_coords = reducer.fit_transform(coord_matrix)
        elif dim_reduction == 'spectral':
            if spectral_affinity == 'precomputed':
                coord_to_idx = {coord: i for i, coord in enumerate(coords)}
                n = len(coords)
                adjacency_matrix = np.zeros((n, n))
                spec_ndim = len(coords[0]) if coords else 0
                for coord in coords:
                    ci = coord_to_idx[coord]
                    for d in range(spec_ndim):
                        neighbor = list(coord)
                        neighbor[d] += 1
                        neighbor = tuple(neighbor)
                        if neighbor in coord_to_idx:
                            adjacency_matrix[ci, coord_to_idx[neighbor]] = 1
                            adjacency_matrix[coord_to_idx[neighbor], ci] = 1
                reducer = SpectralEmbedding(n_components=target_dims, affinity='precomputed', random_state=42)
                reduced_coords = reducer.fit_transform(adjacency_matrix)
            else:
                reducer = SpectralEmbedding(n_components=target_dims, affinity=spectral_affinity,
                                            gamma=spectral_gamma, random_state=42)
                reduced_coords = reducer.fit_transform(coord_matrix)
        else:
            raise ValueError(f"Unknown dim_reduction method: {dim_reduction}. Use 'mds' or 'spectral'")

        coords = [tuple(reduced_coords[i]) for i in range(len(coords))]
        effective_dimensionality = target_dims

        coord_mapping = {original_coords[i]: coords[i] for i in range(len(coords))}
        G_reduced = nx.Graph()
        G_reduced.add_nodes_from(coords)
        orig_set = set(original_coords)
        orig_ndim = len(original_coords[0]) if original_coords else 0
        for coord in original_coords:
            for d in range(orig_ndim):
                neighbor = list(coord)
                neighbor[d] += 1
                neighbor = tuple(neighbor)
                if neighbor in orig_set:
                    G_reduced.add_edge(coord_mapping[coord], coord_mapping[neighbor])
        G = G_reduced
    else:
        effective_dimensionality = lattice.dimensionality
        coord_mapping = {}

    return coords, G, original_coords, coord_mapping, effective_dimensionality


def _build_lattice_graph_data(lattice, coords, original_coords, coord_mapping,
                              effective_dimensionality, nodes, path, shape,
                              path_mode, mute_background, is_tone_lattice):
    valid_coords = set(coords) if lattice.dimensionality <= 3 else set(original_coords)
    highlighted_coords = set()

    if nodes:
        highlighted_coords.update(coord for coord in nodes if coord in valid_coords)
    if path:
        highlighted_coords.update(coord for coord in path if coord in valid_coords)

    if shape is not None and len(shape) > 0:
        if isinstance(shape[0], list):
            lattice_shape_groups = [list(g) for g in shape]
        else:
            lattice_shape_groups = [list(shape)]
        for g in lattice_shape_groups:
            for coord in g:
                c = tuple(coord) if not isinstance(coord, tuple) else coord
                if c in valid_coords:
                    highlighted_coords.add(c)
    else:
        lattice_shape_groups = []

    has_shape = len(lattice_shape_groups) > 0
    use_dimmed = (bool(nodes) or bool(path) or has_shape) and not mute_background

    gen_labels = [str(g) for g in lattice.generators] if is_tone_lattice else []

    return highlighted_coords, lattice_shape_groups, use_dimmed, gen_labels


def _setup_lattice_animation(lattice, coords, G, original_coords, coord_mapping,
                              effective_dimensionality, highlighted_coords,
                              lattice_shape_groups, use_dimmed, mute_background,
                              gen_labels, is_tone_lattice, coord_label,
                              path, path_mode, shape,
                              figsize, node_size, title, path_cmap,
                              dur, arp, strum, direction, amp, kwargs):
    has_shape = len(lattice_shape_groups) > 0
    from klotho.utils.playback._config import get_audio_engine
    engine_name = get_audio_engine()

    svg_threejs_kwargs = dict(
        lattice=lattice, coords=coords, G=G, path=path, nodes=None,
        highlighted_coords=highlighted_coords, coord_mapping=coord_mapping,
        original_coords=original_coords,
        effective_dimensionality=effective_dimensionality,
        use_dimmed=use_dimmed, mute_background=mute_background,
        path_mode=path_mode, figsize=figsize, node_size=node_size,
        title=title, is_tone_lattice=is_tone_lattice,
        coord_label=coord_label, gen_labels=gen_labels,
        path_cmap=path_cmap,
    )

    if path and len(path) > 1:
        preview_config = {
            "dur": kwargs.get("dur", dur) if kwargs else dur,
            "amp": kwargs.get("amp", amp) if kwargs else amp,
            "defName": (kwargs.get("defName") if kwargs else None) or "kl_tri",
            "engine": engine_name,
        }
        from klotho.utils.playback.animation_events import build_path_engine_payload

        ref_freq = 261.63
        freqs = []
        for coord in path:
            try:
                ratio = lattice._coord_to_ratio(coord) if hasattr(lattice, '_coord_to_ratio') else None
                freqs.append(ref_freq * float(ratio) if ratio is not None else ref_freq)
            except Exception:
                freqs.append(ref_freq)
        extra_synth_kwargs = ({k: v for k, v in kwargs.items() if k not in {"pause", "loop", "ring_time"}} if kwargs else None)
        audio_payload = build_path_engine_payload(
            freqs,
            dur,
            engine=engine_name,
            amp=amp,
            extra_pfields=extra_synth_kwargs,
            pause=kwargs.get("pause", 0.0) if kwargs else 0.0,
        )

        if effective_dimensionality <= 2:
            from .._animation import AnimatedLatticeSvgFigure
            from .._renderers.svg_lattice import _svg_lattice_2d
            svg_data = _svg_lattice_2d(**svg_threejs_kwargs, preview_config=preview_config)
            return AnimatedLatticeSvgFigure(
                svg_data=svg_data, audio_payload=audio_payload, dur=dur,
                ring_time=kwargs.get("ring_time", 5) if kwargs else 5,
                loop=kwargs.get("loop", False) if kwargs else False,
            )
        else:
            from .._animation import AnimatedLattice3dFigure
            from .._renderers.threejs_lattice import _threejs_lattice_3d
            threejs_data = _threejs_lattice_3d(**svg_threejs_kwargs, preview_engine=engine_name)
            return AnimatedLattice3dFigure(
                scene_data=threejs_data, audio_payload=audio_payload, dur=dur,
                ring_time=kwargs.get("ring_time", 5) if kwargs else 5,
                loop=kwargs.get("loop", False) if kwargs else False,
            )

    if has_shape:
        preview_config = {
            "dur": kwargs.get("dur", dur) if kwargs else dur,
            "amp": kwargs.get("amp", amp) if kwargs else amp,
            "defName": (kwargs.get("defName") if kwargs else None) or "kl_tri",
            "engine": engine_name,
        }
        from klotho.utils.playback.animation_events import build_shape_engine_payload

        ref_freq = 261.63
        freq_groups = []
        for group in lattice_shape_groups:
            group_freqs = []
            for coord in group:
                try:
                    ratio = lattice._coord_to_ratio(coord) if hasattr(lattice, '_coord_to_ratio') else None
                    group_freqs.append(ref_freq * float(ratio) if ratio is not None else ref_freq)
                except Exception:
                    group_freqs.append(ref_freq)
            freq_groups.append(group_freqs)

        extra_synth_kwargs = ({k: v for k, v in kwargs.items() if k not in {"pause", "loop", "ring_time"}} if kwargs else None)
        audio_payload = build_shape_engine_payload(
            freq_groups,
            dur,
            engine=engine_name,
            arp=arp,
            strum=strum,
            direction=direction,
            amp=amp,
            extra_pfields=extra_synth_kwargs,
            pause=kwargs.get("pause", 0.25) if kwargs else 0.25,
        )

        if effective_dimensionality <= 2:
            from .._animation import AnimatedLatticeShapeFigure
            from .._renderers.svg_lattice import _svg_lattice_2d
            svg_data = _svg_lattice_2d(**svg_threejs_kwargs, shape=lattice_shape_groups, preview_config=preview_config)
            return AnimatedLatticeShapeFigure(
                svg_data=svg_data, audio_payload=audio_payload, dur=dur,
                ring_time=kwargs.get("ring_time", 5) if kwargs else 5,
                loop=kwargs.get("loop", False) if kwargs else False,
            )

    return None


def _plot_lattice(lattice: Lattice, figsize: tuple[float, float] = (12, 12),
                 node_size: float = 8, title: str = None,
                 output_file: str = None,
                 dim_reduction: str = None, target_dims: int = 3,
                 mds_metric: bool = True, mds_max_iter: int = 300,
                 spectral_affinity: str = 'rbf', spectral_gamma: float = None,
                 nodes: list = None, path: list = None, path_mode: str = 'adjacent',
                 mute_background: bool = False, fit=False, pad: int = 1,
                 path_cmap: str = 'viridis',
                 animate: bool = False, dur: float = 0.5,
                 shape: list = None,
                 arp: bool = False, strum: float = 0, direction: str = 'u',
                 amp: float = None, **kwargs):
    """
    Render a Lattice as a 2D SVG or 3D Three.js grid visualization.

    Supports dimensionality reduction for lattices with more than 3
    dimensions.  Coordinate highlighting, path traversal, and animated
    chord / chord-sequence playback are available as overlays.

    Parameters
    ----------
    lattice : Lattice
        Lattice instance to visualize.
    figsize : tuple of float, optional
        Width and height of the figure in inches.
    node_size : float, optional
        Size of the drawn nodes.
    title : str or None, optional
        Plot title.  Auto-generated when ``None``.
    output_file : str or None, optional
        Path to save the figure.
    dim_reduction : str or None, optional
        ``'mds'``, ``'spectral'``, or ``None`` (raises for dim > 3).
    target_dims : int, optional
        Target dimensions after reduction (2 or 3).
    mds_metric : bool, optional
        Use metric MDS when ``True``.
    mds_max_iter : int, optional
        Maximum MDS iterations.
    spectral_affinity : str, optional
        Kernel for spectral embedding.
    spectral_gamma : float or None, optional
        Kernel coefficient for rbf.
    nodes : list of tuple or None, optional
        Coordinate tuples to highlight.
    path : list of tuple or None, optional
        Coordinate tuples defining a traversal path.
    path_mode : str, optional
        ``'adjacent'`` (default) or ``'origin'``.
    mute_background : bool, optional
        Hide non-selected nodes when ``True``.
    fit : bool or str, optional
        Cropping mode when nodes/path are provided.  ``'rect'`` /
        ``True``, ``'square'``, or ``'tight'``.
    pad : int, optional
        Padding around the selection in each dimension when *fit* is
        active.
    path_cmap : str, optional
        Matplotlib colormap for path edge colouring.
    animate : bool, optional
        Return an animated figure when ``True``.
    dur : float, optional
        Seconds between animation steps.
    shape : list or None, optional
        Coordinate tuples (chord) or list of lists (chord sequence).
    arp : bool, optional
        Arpeggiate chord notes sequentially.
    strum : float, optional
        Per-note timing offset (0--1) in a chord.
    direction : str, optional
        ``'u'`` for ascending or ``'d'`` for descending pitch order.

    Returns
    -------
    SvgLatticeData, ThreejsLatticeData, or animated figure
        Renderable figure data.

    Raises
    ------
    ValueError
        If lattice dimensionality exceeds 3 and *dim_reduction* is
        ``None``.
    """
    is_tone_lattice = hasattr(lattice, '_coord_to_ratio')
    coord_label = "Coordinate"
    if is_tone_lattice:
        coord_label = getattr(lattice, "coord_label", "Monzo")

    nodes = _coerce_to_tuples(nodes)
    path = _coerce_to_tuples(path)

    if nodes is not None:
        invalid = [c for c in nodes if c not in lattice]
        if invalid:
            raise ValueError(
                f"The following node coordinates are outside the lattice bounds: {invalid}"
            )
    if path is not None:
        invalid = [c for c in path if c not in lattice]
        if invalid:
            raise ValueError(
                f"The following path coordinates are outside the lattice bounds: {invalid}"
            )

    coords, G, original_coords, coord_mapping, effective_dimensionality = (
        _prepare_lattice_coordinates(
            lattice, nodes, path, path_mode, fit, pad,
            dim_reduction, target_dims,
            mds_metric, mds_max_iter,
            spectral_affinity, spectral_gamma,
        )
    )

    if title is None:
        resolution_str = 'x'.join(str(r) for r in lattice.resolution)
        bipolar_str = "bipolar" if lattice.bipolar else "unipolar"
        if lattice.dimensionality > 3:
            title = f"{lattice.dimensionality}D→{target_dims}D Lattice ({resolution_str}, {bipolar_str}, {dim_reduction})"
        else:
            title = f"{lattice.dimensionality}D Lattice ({resolution_str}, {bipolar_str})"

    highlighted_coords, lattice_shape_groups, use_dimmed, gen_labels = (
        _build_lattice_graph_data(
            lattice, coords, original_coords, coord_mapping,
            effective_dimensionality, nodes, path, shape,
            path_mode, mute_background, is_tone_lattice,
        )
    )

    from .._renderers.svg_lattice import _svg_lattice_2d
    from .._renderers.threejs_lattice import _threejs_lattice_3d
    from klotho.utils.playback._config import get_audio_engine

    svg_threejs_kwargs = dict(
        lattice=lattice, coords=coords, G=G, path=path, nodes=nodes,
        highlighted_coords=highlighted_coords, coord_mapping=coord_mapping,
        original_coords=original_coords,
        effective_dimensionality=effective_dimensionality,
        use_dimmed=use_dimmed, mute_background=mute_background,
        path_mode=path_mode, figsize=figsize, node_size=node_size,
        title=title, is_tone_lattice=is_tone_lattice,
        coord_label=coord_label, gen_labels=gen_labels,
        path_cmap=path_cmap,
    )
    preview_engine = get_audio_engine()

    if not animate:
        if effective_dimensionality <= 2:
            return _svg_lattice_2d(**svg_threejs_kwargs, shape=lattice_shape_groups if lattice_shape_groups else None)
        else:
            return _threejs_lattice_3d(**svg_threejs_kwargs, preview_engine=preview_engine)

    animated_fig = _setup_lattice_animation(
        lattice, coords, G, original_coords, coord_mapping,
        effective_dimensionality, highlighted_coords,
        lattice_shape_groups, use_dimmed, mute_background,
        gen_labels, is_tone_lattice, coord_label,
        path, path_mode, shape,
        figsize, node_size, title, path_cmap,
        dur, arp, strum, direction, amp, kwargs,
    )

    if animated_fig is not None:
        return animated_fig

    if effective_dimensionality <= 2:
        return _svg_lattice_2d(**svg_threejs_kwargs, shape=lattice_shape_groups if lattice_shape_groups else None)
    return _threejs_lattice_3d(**svg_threejs_kwargs, preview_engine=preview_engine)



from .frequency_conversion import pitchclass_to_freq, freq_to_pitchclass
from fractions import Fraction
from typing import Tuple, Union

__all__ = [
    'partial_to_fundamental',
    'first_equave'
]

def partial_to_fundamental(pitchclass: str, octave: int = 4, partial: int = 1, cent_offset: float = 0.0) -> Tuple[str, float]:
    """
    Calculate the fundamental from a pitch and its partial number.

    Given a pitch that represents a specific partial of an unknown
    fundamental, this function back-calculates the fundamental frequency
    and returns its pitch-class information.

    Parameters
    ----------
    pitchclass : str
        Pitch class name (e.g., ``"A"``, ``"C#"``).
    octave : int, optional
        Octave number. Default is 4.
    partial : int, optional
        Partial number (non-zero). Negative values indicate undertones.
        Default is 1.
    cent_offset : float, optional
        Microtonal offset in cents. Default is 0.0.

    Returns
    -------
    namedtuple
        A named tuple with fields ``pitchclass``, ``octave``, and
        ``cents_offset`` for the fundamental.

    Raises
    ------
    ValueError
        If *partial* is zero.
    """
    if partial == 0:
        raise ValueError("Partial number cannot be zero")

    freq = pitchclass_to_freq(pitchclass, octave, cent_offset)
    # For negative partials (undertones), multiply by |p| instead of dividing
    fundamental_freq = freq * abs(partial) if partial < 0 else freq / partial
    return freq_to_pitchclass(fundamental_freq)

def first_equave(harmonic: Union[int, float, Fraction], equave: Union[int, float, Fraction, str] = 2, max_equave: Union[int, float, Fraction, str] = None):
  """
  Return the first equave register in which a harmonic number appears.

  For example, harmonics 2 and 3 first appear in equave 1 (range 2--4),
  harmonic 5 first appears in equave 2 (range 4--8), etc.

  Parameters
  ----------
  harmonic : int, float, or Fraction
      The harmonic number.
  equave : int, float, Fraction, or str, optional
      Interval of equivalence. Default is 2 (octave).
  max_equave : int, float, Fraction, str, or None, optional
      Maximum equave register to search. None means unbounded.

  Returns
  -------
  int or None
      The equave register number, or None if not found within
      *max_equave*.
  """
  equave = Fraction(equave)
  max_equave = Fraction(max_equave) if max_equave is not None else None
  n_equave = 0
  while max_equave is None or n_equave <= max_equave:
    if harmonic <= equave ** n_equave:
      return n_equave
    n_equave += 1
  return None



import copy

from klotho.tonos import Pitch
from klotho.tonos.pitch.pitch_collections import PitchCollectionBase
from klotho.tonos.chords.chord import Chord, Voicing, ChordSequence
from klotho.tonos.scales.scale import Scale
from klotho.tonos.systems.harmonic_trees import Spectrum, HarmonicTree
from klotho.chronos.rhythm_trees.rhythm_tree import RhythmTree
from klotho.chronos.temporal_units.temporal import TemporalUnit, TemporalUnitSequence, TemporalBlock
from klotho.thetos.composition.compositional import CompositionalUnit
from klotho.thetos.instruments.tone import ToneInstrument
from klotho.utils.playback._amplitude import single_voice_amplitude, compute_voice_amplitudes
from klotho.utils.playback._converter_base import (
    DEFAULT_NOTE_DURATION, DEFAULT_CHORD_DURATION,
    DEFAULT_SPECTRUM_DURATION, DEFAULT_DRUM_FREQ,
    KNOWN_KWARGS,
    _get_addressed_collection, _merge_pfields,
    scale_pitch_sequence, extract_convert_kwargs, lower_event_ir_to_voice_events, iter_group_sequence,
)


def _payload(events, instruments=None):
    return {"events": events, "instruments": instruments or {}}


def _deep_merge(base, override):
    merged = copy.deepcopy(base)
    for key, value in override.items():
        if isinstance(value, dict) and isinstance(merged.get(key), dict):
            merged[key] = _deep_merge(merged[key], value)
        else:
            merged[key] = value
    return merged


def _normalize_event_pfields(pfields):
    if 'vel' not in pfields and 'amp' in pfields:
        pfields['vel'] = pfields['amp']
    if 'amp' not in pfields and 'vel' in pfields:
        pfields['amp'] = pfields['vel']
    if 'freq' not in pfields:
        pfields['freq'] = 440.0
    return pfields


def _build_seq_events(pitches, start, instrument, amp=None, per_voice_dur=None,
                      total_dur=None, pause=0.0, extra_pfields=None):
    events = []
    n = len(pitches)
    if n == 0:
        return events

    if total_dur is not None:
        voice_dur = total_dur / n
    elif per_voice_dur is not None:
        voice_dur = per_voice_dur
    else:
        voice_dur = DEFAULT_NOTE_DURATION

    cursor = start
    for i, pitch in enumerate(pitches):
        pf = {
            "freq": pitch.freq,
            "vel": single_voice_amplitude(pitch.freq, amp),
        }
        pf = _merge_pfields(pf, extra_pfields)
        events.append({
            "start": cursor,
            "duration": voice_dur,
            "instrument": instrument,
            "pfields": pf,
        })
        cursor += voice_dur + max(0.0, pause)
    return events


def _build_chord_events(pitches, start, dur, strum, instrument, amp=None,
                        dur_factor=1.0, extra_pfields=None):
    events = []
    num_notes = len(pitches)
    if num_notes == 0:
        return events

    freqs = [p.freq for p in pitches]
    voice_amps = compute_voice_amplitudes(freqs, amp)
    strum = max(0, min(1, strum))

    for i, pitch in enumerate(pitches):
        start_offset = (strum * dur * i) / num_notes if num_notes > 1 else 0
        pf = {
            "freq": pitch.freq,
            "vel": voice_amps[i],
        }
        pf = _merge_pfields(pf, extra_pfields)
        events.append({
            "start": start + start_offset,
            "duration": (dur * dur_factor) - start_offset,
            "instrument": instrument,
            "pfields": pf,
        })
    return events


def compositional_unit_to_events(obj, extra_pfields=None, animation=False):
    events = []
    instruments = {}
    inst_id_map = {}

    leaf_nodes = obj._rt.leaf_nodes if animation else None
    node_to_step = ({nid: idx for idx, nid in enumerate(leaf_nodes)}
                    if animation else None)

    time_offset = 0
    if animation:
        time_offset = min(ev.start for ev in obj if not ev.is_rest) if any(not ev.is_rest for ev in obj) else 0

    for event in obj:
        step_idx = node_to_step.get(event.node_id, None) if animation else None
        start_time = event.start - time_offset if animation else event.start

        if event.is_rest:
            if animation:
                events.append({
                    "start": start_time,
                    "duration": abs(event.duration),
                    "instrument": "__rest__",
                    "pfields": {},
                    "_stepIndex": step_idx,
                })
            continue

        instrument = obj.get_instrument(event.node_id)
        if instrument is None:
            instrument = ToneInstrument('Synth', 'Synth')
        elif isinstance(instrument, str):
            instrument = ToneInstrument(instrument, instrument)
        elif not isinstance(instrument, ToneInstrument):
            if animation:
                events.append({
                    "start": start_time,
                    "duration": abs(event.duration),
                    "instrument": "__rest__",
                    "pfields": {},
                    "_stepIndex": step_idx,
                })
            continue

        inst_identity = id(instrument)
        if inst_identity not in inst_id_map:
            key = instrument.name
            if key in instruments:
                existing = instruments[key]
                if existing['tonejs_class'] != instrument.tonejs_class or existing['preset'] != instrument.pfields:
                    raise ValueError(
                        f"Instrument name '{key}' is used by two different instruments. "
                        f"Use distinct names."
                    )
            else:
                instruments[key] = {
                    'tonejs_class': instrument.tonejs_class,
                    'preset': instrument.pfields
                }
            inst_id_map[inst_identity] = key

        routing_key = inst_id_map[inst_identity]
        expanded_voices = lower_event_ir_to_voice_events(event, step_index=step_idx)
        for expanded_voice in expanded_voices:
            pfields = {k: v for k, v in expanded_voice["pfields"].items()
                       if k != 'group'}
            pfields = _normalize_event_pfields(pfields)
            default_pfields = instruments[routing_key].get('preset', {})
            effective_pfields = _deep_merge(default_pfields, pfields)
            effective_pfields = _merge_pfields(effective_pfields, extra_pfields)

            voice_start = expanded_voice["start"] - time_offset if animation else expanded_voice["start"]
            ev_data = {
                "start": voice_start,
                "duration": expanded_voice["duration"],
                "instrument": routing_key,
                "pfields": effective_pfields,
                "_polyGroupId": expanded_voice["poly_group_id"],
                "_logicalStepId": expanded_voice["logical_step_id"],
                "_polyVoiceIndex": expanded_voice["poly_voice_index"],
                "_polyVoiceCount": expanded_voice["poly_voice_count"],
                "_polyLeader": expanded_voice["poly_is_leader"],
                "_animate": expanded_voice["animate"],
            }
            if animation:
                ev_data["_stepIndex"] = step_idx
            events.append(ev_data)

    events.sort(key=lambda ev: ev["start"])
    return _payload(events, instruments)


def compositional_unit_to_animation_events(obj, extra_pfields=None):
    return compositional_unit_to_events(obj, extra_pfields=extra_pfields, animation=True)


def pitch_to_events(pitch, duration=None, amp=None, extra_pfields=None):
    dur = duration if duration is not None else 1.0
    pf = {
        "freq": pitch.freq,
        "vel": single_voice_amplitude(pitch.freq, amp),
    }
    pf = _merge_pfields(pf, extra_pfields)
    return _payload([{
        "start": 0.0,
        "duration": dur,
        "instrument": "synth",
        "pfields": pf,
    }])


def pitch_collection_to_events(obj, duration=None, mode="seq", arp=False, strum=0, direction='u',
                               amp=None, pause=0.0, extra_pfields=None):
    addressed = _get_addressed_collection(obj)
    pitches = [addressed[i] for i in range(len(addressed))]

    if mode == "chord":
        pitches = sorted(pitches, key=lambda p: p.freq)
        if direction.lower() == 'd':
            pitches = list(reversed(pitches))

        if arp:
            dur = duration if duration is not None else DEFAULT_CHORD_DURATION
            return _payload(_build_seq_events(pitches, 0, "synth", amp=amp,
                                              total_dur=dur, pause=0.0, extra_pfields=extra_pfields))
        else:
            dur = duration if duration is not None else DEFAULT_CHORD_DURATION
            return _payload(_build_chord_events(pitches, 0, dur, strum, "synth",
                                                amp=amp, extra_pfields=extra_pfields))
    else:
        dur = duration if duration is not None else DEFAULT_NOTE_DURATION
        return _payload(_build_seq_events(pitches, 0, "synth", amp=amp,
                                          per_voice_dur=dur, pause=pause, extra_pfields=extra_pfields))


def scale_to_events(obj, duration=None, equaves=1, amp=None, pause=0.0, extra_pfields=None):
    dur = duration if duration is not None else DEFAULT_NOTE_DURATION
    all_pitches = scale_pitch_sequence(obj, equaves)
    return _payload(_build_seq_events(all_pitches, 0, "synth", amp=amp,
                                      per_voice_dur=dur, pause=pause, extra_pfields=extra_pfields))


def chord_to_events(obj, duration=None, arp=False, strum=0, direction='u',
                    amp=None, extra_pfields=None):
    addressed = _get_addressed_collection(obj)
    pitches = [addressed[i] for i in range(len(addressed))]

    if direction.lower() == 'd':
        pitches = list(reversed(pitches))

    if arp:
        dur = duration if duration is not None else DEFAULT_CHORD_DURATION
        return _payload(_build_seq_events(pitches, 0, "synth", amp=amp,
                                          total_dur=dur, pause=0.0, extra_pfields=extra_pfields))
    else:
        dur = duration if duration is not None else DEFAULT_CHORD_DURATION
        return _payload(_build_chord_events(pitches, 0, dur, strum, "synth",
                                            amp=amp, extra_pfields=extra_pfields))


def chord_sequence_to_events(obj, duration=None, arp=False, strum=0, direction='u',
                             amp=None, pause=0.25, extra_pfields=None):
    events = []
    dur = duration if duration is not None else DEFAULT_CHORD_DURATION
    groups = []
    for chord in obj:
        addressed = _get_addressed_collection(chord)
        groups.append([addressed[i] for i in range(len(addressed))])
    group_voice_amps = [
        compute_voice_amplitudes([p.freq for p in group], amp)
        for group in groups
    ]

    if arp:
        for _, _, start, voice_dur, pitch in iter_group_sequence(groups, dur, arp=True, direction=direction, pause=pause):
            pf = {
                "freq": pitch.freq,
                "vel": single_voice_amplitude(pitch.freq, amp),
            }
            pf = _merge_pfields(pf, extra_pfields)
            events.append({
                "start": start,
                "duration": voice_dur,
                "instrument": "synth",
                "pfields": pf,
            })
    else:
        for gi, vi, start, voice_dur, pitch in iter_group_sequence(groups, dur, arp=False, strum=strum, direction=direction, pause=pause):
            pf = {
                "freq": pitch.freq,
                "vel": group_voice_amps[gi][vi],
            }
            pf = _merge_pfields(pf, extra_pfields)
            events.append({
                "start": start,
                "duration": voice_dur,
                "instrument": "synth",
                "pfields": pf,
            })

    return _payload(events)


def spectrum_to_events(obj, duration=None, arp=False, strum=0, direction='u',
                       amp=None, extra_pfields=None):
    pitches = [row['pitch'] for _, row in obj.data.iterrows()]

    if direction.lower() == 'd':
        pitches = list(reversed(pitches))

    target = amp if amp is not None else 0.4
    if arp:
        dur = duration if duration is not None else DEFAULT_SPECTRUM_DURATION
        return _payload(_build_seq_events(pitches, 0, "sine", amp=target,
                                          total_dur=dur, pause=0.0, extra_pfields=extra_pfields))
    else:
        dur = duration if duration is not None else DEFAULT_SPECTRUM_DURATION
        return _payload(_build_chord_events(pitches, 0, dur, strum, "sine",
                                            amp=target, extra_pfields=extra_pfields))


def temporal_unit_to_events(obj, use_absolute_time=False, amp=None, extra_pfields=None,
                            animation=False):
    events = []
    target = amp if amp is not None else 0.85

    leaf_nodes = obj._rt.leaf_nodes if animation else None
    node_to_step = ({nid: idx for idx, nid in enumerate(leaf_nodes)}
                    if animation else None)

    if use_absolute_time:
        time_offset = 0
    else:
        time_offset = min(chronon.start for chronon in obj if not chronon.is_rest) if any(not chronon.is_rest for chronon in obj) else 0

    for chronon in obj:
        start_time = chronon.start - time_offset
        duration = abs(chronon.duration)
        step_idx = node_to_step.get(chronon.node_id, None) if animation else None

        if chronon.is_rest:
            if animation:
                events.append({
                    "start": start_time,
                    "duration": duration,
                    "instrument": "__rest__",
                    "pfields": {},
                    "_stepIndex": step_idx,
                })
            continue

        pf = {
            "freq": DEFAULT_DRUM_FREQ,
            "vel": target,
        }
        pf = _merge_pfields(pf, extra_pfields)
        ev_data = {
            "start": start_time,
            "duration": duration,
            "instrument": "membrane",
            "pfields": pf,
        }
        if animation:
            ev_data["_stepIndex"] = step_idx
        events.append(ev_data)

    if animation:
        events.sort(key=lambda ev: ev["start"])
    return _payload(events)


def temporal_unit_to_animation_events(obj, use_absolute_time=False, amp=None, extra_pfields=None):
    return temporal_unit_to_events(obj, use_absolute_time=use_absolute_time, amp=amp,
                                  extra_pfields=extra_pfields, animation=True)


def rhythm_tree_to_events(obj, beat=None, bpm=None, amp=None, extra_pfields=None):
    temporal_unit = TemporalUnit.from_rt(obj, beat=beat, bpm=bpm)
    return temporal_unit_to_events(temporal_unit, use_absolute_time=False,
                                  amp=amp, extra_pfields=extra_pfields)


def rhythm_tree_to_animation_events(obj, beat=None, bpm=None, amp=None, extra_pfields=None):
    temporal_unit = TemporalUnit.from_rt(obj, beat=beat, bpm=bpm)
    return temporal_unit_to_animation_events(temporal_unit, use_absolute_time=False,
                                            amp=amp, extra_pfields=extra_pfields)


def _merge_sub_payload(target_events, target_instruments, sub_payload, time_offset):
    for ev in sub_payload["events"]:
        ev["start"] -= time_offset
    target_events.extend(sub_payload["events"])
    for key, val in sub_payload["instruments"].items():
        if key in target_instruments:
            existing = target_instruments[key]
            if existing['tonejs_class'] != val['tonejs_class'] or existing['preset'] != val['preset']:
                raise ValueError(
                    f"Instrument name '{key}' is used by two different instruments. "
                    f"Use distinct names."
                )
        else:
            target_instruments[key] = val


def temporal_sequence_to_events(obj, extra_pfields=None):
    events = []
    instruments = {}
    seq_offset = obj.offset

    for unit in obj:
        if isinstance(unit, CompositionalUnit):
            _merge_sub_payload(events, instruments, compositional_unit_to_events(unit, extra_pfields=None), seq_offset)
        elif isinstance(unit, TemporalUnit):
            _merge_sub_payload(events, instruments, temporal_unit_to_events(unit, use_absolute_time=True, extra_pfields=extra_pfields), seq_offset)
        elif isinstance(unit, TemporalUnitSequence):
            _merge_sub_payload(events, instruments, temporal_sequence_to_events(unit, extra_pfields=extra_pfields), seq_offset)
        elif isinstance(unit, TemporalBlock):
            _merge_sub_payload(events, instruments, temporal_block_to_events(unit, extra_pfields=extra_pfields), seq_offset)

    events.sort(key=lambda ev: ev["start"])
    return _payload(events, instruments)


def temporal_block_to_events(obj, extra_pfields=None):
    events = []
    instruments = {}
    block_offset = obj.offset

    for row in obj:
        if isinstance(row, CompositionalUnit):
            _merge_sub_payload(events, instruments, compositional_unit_to_events(row, extra_pfields=None), block_offset)
        elif isinstance(row, TemporalUnit):
            _merge_sub_payload(events, instruments, temporal_unit_to_events(row, use_absolute_time=True, extra_pfields=extra_pfields), block_offset)
        elif isinstance(row, TemporalUnitSequence):
            _merge_sub_payload(events, instruments, temporal_sequence_to_events(row, extra_pfields=extra_pfields), block_offset + row.offset)
        elif isinstance(row, TemporalBlock):
            _merge_sub_payload(events, instruments, temporal_block_to_events(row, extra_pfields=extra_pfields), block_offset + row.offset)

    events.sort(key=lambda ev: ev["start"])
    return _payload(events, instruments)


def convert_to_events(obj, **kwargs):
    kw = extract_convert_kwargs(kwargs)
    duration = kw['duration']
    arp = kw['arp']
    mode = kw['mode']
    strum = kw['strum']
    direction = kw['direction']
    equaves = kw['equaves']
    beat = kw['beat']
    bpm = kw['bpm']
    amp = kw['amp']
    pause = kw['pause']
    extra_pfields = kw['extra_pfields']

    if isinstance(obj, Pitch):
        return pitch_to_events(obj, duration=duration, amp=amp, extra_pfields=extra_pfields)

    if isinstance(obj, Spectrum):
        return spectrum_to_events(obj, duration=duration, arp=arp, strum=strum, direction=direction,
                                  amp=amp, extra_pfields=extra_pfields)

    if isinstance(obj, HarmonicTree):
        spectrum = Spectrum(Pitch("C4"), list(obj.partials) if hasattr(obj, 'partials') else [1, 2, 3, 4, 5])
        return spectrum_to_events(spectrum, duration=duration, arp=arp, strum=strum, direction=direction,
                                  amp=amp, extra_pfields=extra_pfields)

    if isinstance(obj, RhythmTree):
        return rhythm_tree_to_events(obj, beat=beat, bpm=bpm, amp=amp, extra_pfields=extra_pfields)

    if isinstance(obj, TemporalUnitSequence):
        return temporal_sequence_to_events(obj, extra_pfields=extra_pfields)

    if isinstance(obj, TemporalBlock):
        return temporal_block_to_events(obj, extra_pfields=extra_pfields)

    if isinstance(obj, CompositionalUnit):
        return compositional_unit_to_events(obj, extra_pfields=None)

    if isinstance(obj, TemporalUnit):
        return temporal_unit_to_events(obj, amp=amp, extra_pfields=extra_pfields)

    if isinstance(obj, ChordSequence):
        return chord_sequence_to_events(obj, duration=duration, arp=arp, strum=strum, direction=direction,
                                        amp=amp, pause=(0.25 if pause is None else pause), extra_pfields=extra_pfields)

    if isinstance(obj, Scale):
        return scale_to_events(obj, duration=duration, equaves=equaves, amp=amp,
                               pause=(0.0 if pause is None else pause), extra_pfields=extra_pfields)

    if isinstance(obj, (Chord, Voicing)):
        return chord_to_events(obj, duration=duration, arp=arp, strum=strum, direction=direction,
                               amp=amp, extra_pfields=extra_pfields)

    if isinstance(obj, PitchCollectionBase):
        effective_mode = mode if mode else "sequential"
        if effective_mode == "chord":
            return pitch_collection_to_events(obj, duration=duration, mode="chord", arp=arp, strum=strum,
                                              direction=direction, amp=amp,
                                              pause=0.0,
                                              extra_pfields=extra_pfields)
        return pitch_collection_to_events(obj, duration=duration, mode="sequential",
                                          amp=amp, pause=(0.0 if pause is None else pause),
                                          extra_pfields=extra_pfields)

    raise TypeError(f"Unsupported object type: {type(obj)}")



import json
import uuid
import base64
from pathlib import Path

from IPython.display import HTML, display

from klotho.utils.playback.supersonic.cdn import supersonic_config
from klotho.utils.playback._helpers import convert_numpy_types
from klotho.utils.playback.supersonic._js_fragments import (
    ss_init_js, draw_scheduler_js, scheduler_core_js, scheduler_score_js,
    synthdef_registry_merge_js, control_bar_html,
)

SYNTHDEFS_DIR = Path(__file__).parent / "assets" / "synthdefs"
_WIDGET_JS_PATH = Path(__file__).parent / "_engine_widget.js"
_WIDGET_JS_TEMPLATE = None
_ALL_SYNTHDEF_ASSETS = None

_INFRA_SYNTHDEFS = frozenset({'__busRouter', '__busRouterMonitor', '__chainLimiter'})


def _load_all_synthdef_assets():
    global _ALL_SYNTHDEF_ASSETS
    if _ALL_SYNTHDEF_ASSETS is None:
        assets = {}
        if SYNTHDEFS_DIR.exists():
            for path in SYNTHDEFS_DIR.glob("*.scsyndef"):
                assets[path.stem] = base64.b64encode(path.read_bytes()).decode("ascii")
        if "default" not in assets and "kl_tri" in assets:
            assets["default"] = assets["kl_tri"]
        _ALL_SYNTHDEF_ASSETS = assets
    return _ALL_SYNTHDEF_ASSETS


def _filter_synthdef_assets(all_assets, needed):
    filtered = {}
    for name in needed | _INFRA_SYNTHDEFS:
        if name in all_assets:
            filtered[name] = all_assets[name]
    if "default" not in filtered and "kl_tri" in all_assets:
        filtered["default"] = all_assets["kl_tri"]
    return filtered


def _load_widget_template():
    global _WIDGET_JS_TEMPLATE
    if _WIDGET_JS_TEMPLATE is None:
        _WIDGET_JS_TEMPLATE = _WIDGET_JS_PATH.read_text()
    return _WIDGET_JS_TEMPLATE


class SuperSonicEngine:
    def __init__(self, events, meta=None, control_data=None, ring_time=5):
        self.events = convert_numpy_types(events)
        self.meta = meta or {}
        self.control_data = control_data or {"buffer": None, "blockSize": 512, "descriptors": []}
        self.ring_time = ring_time
        self.widget_id = f"klotho_ss_{uuid.uuid4().hex[:8]}"
        from klotho.utils.playback._sc_validate import validate_sc_events, validate_sc_meta
        validate_sc_events(self.events)
        if self.meta:
            validate_sc_meta(self.meta)
        self._needed = self._needed_synthdefs() | _INFRA_SYNTHDEFS
        self.synthdef_assets = _filter_synthdef_assets(_load_all_synthdef_assets(), self._needed)
        self._is_score = bool(self.meta.get("groups") or self.meta.get("inserts"))

    def _needed_synthdefs(self):
        names = set()
        for ev in self.events:
            if ev.get("type") == "new" and ev.get("defName"):
                name = ev["defName"]
                if name != "__rest__":
                    names.add(name)
        for track_inserts in self.meta.get("inserts", {}).values():
            for ins in track_inserts:
                dn = ins.get("defName")
                if dn:
                    names.add(dn)
        if self.control_data.get("descriptors"):
            names.add("__klEnvCtrl")
        return names

    def _serialize_control_data(self):
        cd = self.control_data
        result = {"blockSize": cd.get("blockSize", 512), "descriptors": cd.get("descriptors", []), "bufferB64": None, "numFrames": 0}
        buf = cd.get("buffer")
        if buf is not None:
            result["bufferB64"] = base64.b64encode(buf.tobytes()).decode("ascii")
            result["numFrames"] = len(buf)
        return result

    def _generate_html(self):
        events_json = json.dumps(self.events)
        synthdef_assets_json = json.dumps(self.synthdef_assets)
        needed_json = json.dumps(list(self._needed))
        config_json = json.dumps(supersonic_config())
        wid = self.widget_id
        meta_json = json.dumps(self.meta)
        control_data_json = json.dumps(self._serialize_control_data())

        widget_js = (_load_widget_template()
                     .replace('__WID__', wid)
                     .replace('__EVENTS_JSON__', events_json)
                     .replace('__NEEDED_JSON__', needed_json)
                     .replace('__SS_CONFIG_JSON__', config_json)
                     .replace('__META_JSON__', meta_json)
                     .replace('__CONTROL_DATA_JSON__', control_data_json)
                     .replace('__RING_TIME__', str(self.ring_time)))

        score_js = scheduler_score_js() if self._is_score else ""

        bar_html = control_bar_html(wid, show_status=True)

        html = f'''
{bar_html}

<script>
{ss_init_js()}
{draw_scheduler_js()}
{scheduler_core_js()}
{score_js}
{synthdef_registry_merge_js(synthdef_assets_json)}
{widget_js}
</script>
'''
        return html

    def display(self):
        html = self._generate_html()
        return display(HTML(html))

    def _repr_html_(self):
        return self._generate_html()



from klotho.chronos.rhythm_trees import RhythmTree


def _plot_rt(rt: RhythmTree, layout: str = 'containers', figsize: tuple[float, float] | None = None, 
            invert: bool = True, output_file: str | None = None, 
            attributes: list[str] | None = None, vertical_lines: bool = True, 
            barlines: bool = True, barline_color: str = '#666666', 
            subdivision_line_color: str = '#aaaaaa',
            animate: bool = False, dur: float = 0.5,
            audio_source=None,
            beat=None, bpm=None,
            glow: bool = False, amp: float = None, **kwargs) -> None:
    """
    Render a RhythmTree using one of several SVG layout modes.

    Parameters
    ----------
    rt : RhythmTree
        RhythmTree instance to visualize.
    layout : str, optional
        Layout mode: ``'containers'`` (nested bars, default),
        ``'ratios'`` (leaf-duration bar), or ``'tree'`` (node-link
        diagram).
    figsize : tuple of float or None, optional
        Width and height in inches.  Defaults vary by layout.
    invert : bool, optional
        When ``True``, places the root at the top.
    output_file : str or None, optional
        Path to save the visualization.
    attributes : list of str or None, optional
        Node attributes to display (tree layout only).
    vertical_lines : bool, optional
        Show subdivision guide lines (containers layout).
    barlines : bool, optional
        Show barlines for multi-span trees (containers layout).
    barline_color : str, optional
        CSS colour for barlines.
    subdivision_line_color : str, optional
        CSS colour for subdivision guide lines.
    animate : bool, optional
        Return an animated figure with audio playback controls.
    dur : float, optional
        Seconds between animation steps.
    audio_source : TemporalUnit or CompositionalUnit or None, optional
        Source of real-time audio data for the animation.
    beat : optional
        Beat value forwarded to the animation event converter.
    bpm : optional
        Tempo value forwarded to the animation event converter.
    glow : bool, optional
        Enable halo glow effect on the active leaf during animation.

    Returns
    -------
    SvgRTData or AnimatedRTSvgFigure
        Renderable SVG data or animated figure.
    """
    from .._renderers.svg_rt import (
        _svg_rt_ratios, _svg_rt_containers, _svg_rt_tree,
    )

    if layout == 'tree':
        if figsize is None:
            figsize = (11, 2)
        svg_data = _svg_rt_tree(rt, attributes=attributes, figsize=figsize, invert=invert,
                                audio_source=audio_source)
    elif layout == 'ratios':
        if figsize is None:
            figsize = (11, 0.5)
        svg_data = _svg_rt_ratios(rt, figsize=figsize, audio_source=audio_source)
    elif layout == 'containers':
        if figsize is None:
            figsize = (11, 2)
        svg_data = _svg_rt_containers(rt, figsize=figsize, invert=invert,
                                      vertical_lines=vertical_lines, barlines=barlines,
                                      barline_color=barline_color,
                                      subdivision_line_color=subdivision_line_color,
                                      audio_source=audio_source)
    else:
        raise ValueError(f"Unknown layout: {layout}. Choose 'tree', 'containers', or 'ratios'.")

    if not animate:
        return svg_data

    from .._animation import AnimatedRTSvgFigure
    from klotho.utils.playback._config import get_audio_engine
    from klotho.chronos.temporal_units.temporal import TemporalUnit as _TemporalUnit
    from klotho.thetos.composition.compositional import CompositionalUnit as _CompositionalUnit

    engine = get_audio_engine()

    if engine == "supersonic":
        from klotho.utils.playback.supersonic.converters import (
            rhythm_tree_to_sc_animation_events,
            temporal_unit_to_sc_animation_events,
            compositional_unit_to_sc_animation_events,
        )
        if audio_source is not None:
            if isinstance(audio_source, _CompositionalUnit):
                audio_payload = compositional_unit_to_sc_animation_events(audio_source)
            elif isinstance(audio_source, _TemporalUnit):
                audio_payload = temporal_unit_to_sc_animation_events(audio_source, amp=amp)
            else:
                audio_payload = rhythm_tree_to_sc_animation_events(rt, beat=beat, bpm=bpm, amp=amp)
        else:
            audio_payload = rhythm_tree_to_sc_animation_events(rt, beat=beat, bpm=bpm, amp=amp)
    else:
        from klotho.utils.playback.tonejs.converters import (
            rhythm_tree_to_animation_events,
            temporal_unit_to_animation_events,
            compositional_unit_to_animation_events,
        )
        if audio_source is not None:
            if isinstance(audio_source, _CompositionalUnit):
                audio_payload = compositional_unit_to_animation_events(audio_source)
            elif isinstance(audio_source, _TemporalUnit):
                audio_payload = temporal_unit_to_animation_events(audio_source, amp=amp)
            else:
                audio_payload = rhythm_tree_to_animation_events(rt, beat=beat, bpm=bpm, amp=amp)
        else:
            audio_payload = rhythm_tree_to_animation_events(rt, beat=beat, bpm=bpm, amp=amp)

    return AnimatedRTSvgFigure(
        svg_data=svg_data,
        audio_payload=audio_payload,
        dur=dur,
        glow=glow,
        engine=engine,
        ring_time=kwargs.get("ring_time", 5),
        loop=kwargs.get("loop", False),
    )




# ------------------------------------------------------------------------------------
# Klotho/klotho/chronos/rhythm_trees/rt.py
# ------------------------------------------------------------------------------------
"""
Rhythm trees.

A rhythm tree (RT) is a list representing a rhythmic structure. This list
is organized hierarchically in sub-lists, just as time is organized in
measures, time signatures, pulses and rhythmic elements in traditional
notation.

See: https://support.ircam.fr/docs/om/om6-manual/co/RT.html
"""
from fractions import Fraction
from typing import Union, Tuple
from tabulate import tabulate

from klotho.topos.graphs import Tree, Group, format_subdivisions
from .meas import Meas
from .algorithms import sum_proportions, measure_complexity, ratios_to_subdivs
from ..utils.beat import calc_onsets


class RhythmTree(Tree):
    _node_value_attr = 'proportion'
    '''
    A rhythm tree is a list representing a rhythmic structure. This list is organized 
    hierarchically in sub lists, just as time is organized in measures, time signatures, 
    pulses and rhythmic elements in the traditional notation.

    Traditionally, rhythm is broken up into several data : meter, measure(s) and duration(s). 
    Rhythm trees must enclose these information in lists and sub list.

    This elementary rhythm:

    [1/4, 1/4, 1/4, 1/4] --> (four 1/4-notes in 4/4 time)

    can be expressed as follows :

    ( ? ( (4//4 (1 1 1 1) ) ) )

    A tree structure can be reduced to a list : (D (S)).


    >> Main Components : Duration and Subdivisions

    D = a duration , or number of measures : ( ? ) or a number ( n ).
    When D = ?, OM calculates the duration.
    By default, this duration is equal to 1.

    S = subdivisions (S) of this duration, that is a time signature and rhythmic proportions.
    Time signature = n // n   or ( n n ).
    It must be specified at each new measure, even if it remains unchanged.

    Rhythm = proportions : ( n n n n )

    see: https://support.ircam.fr/docs/om/om6-manual/co/RT1.html
    '''
    def __init__(self, 
                 span:int                      = 1,
                 meas:Union[Meas,Fraction,str] = '1/1',
                 subdivisions:Tuple            = (1,1)):
        """
        Parameters
        ----------
        span : int, optional
            The number of measures the tree spans. Default is 1.
        meas : Meas, Fraction, or str, optional
            The time signature. Default is ``'1/1'``.
        subdivisions : tuple, optional
            The proportional subdivisions of the measure. Elements may
            be integers or nested ``(D, S)`` tuples. Default is ``(1, 1)``.
        """
        casted = self._cast_subdivs(subdivisions)
        super().__init__(Meas(meas).numerator * span, casted)
        
        self._meta['span'] = span
        self._meta['meas'] = str(Meas(meas))
        self._meta['type'] = None
        self._list = Group((Meas(meas).numerator * span, casted))
        
        self._evaluate()

    @classmethod
    def from_tree(cls, tree:Tree, span:int = 1):
        """
        Construct a ``RhythmTree`` from an existing :class:`Tree`.

        Parameters
        ----------
        tree : Tree
            A tree whose root node has a ``'metric_duration'`` attribute.
        span : int, optional
            The number of measures. Default is 1.

        Returns
        -------
        RhythmTree
        """
        return cls(span = span, meas = Meas(tree[tree.root]['metric_duration']), subdivisions = tree.group.S)
    
    @classmethod
    def from_ratios(cls, ratios:Tuple[Fraction, float, str], span:int = 1):
        """
        Construct a ``RhythmTree`` from a flat sequence of fractional ratios.

        The ratios are converted to integer subdivisions and the time
        signature is inferred from the sum of absolute ratios.

        Parameters
        ----------
        ratios : tuple of Fraction, float, or str
            The proportional ratios for each leaf.
        span : int, optional
            The number of measures. Default is 1.

        Returns
        -------
        RhythmTree
        """
        ratios = tuple(Fraction(r) for r in ratios)
        S = ratios_to_subdivs(ratios)
        meas = Meas(sum(abs(r) for r in ratios))
        return cls(span = span, meas = meas, subdivisions = S)

    @property
    def span(self):
        """
        The number of measures this tree spans.

        Returns
        -------
        int
        """
        return self._meta['span']

    @property
    def meas(self):
        """
        The time signature of this tree.

        Returns
        -------
        Meas
        """
        return Meas(self._meta['meas'])

    @property
    def subdivisions(self):
        """
        The proportional subdivisions (S part) of this tree.

        Returns
        -------
        tuple
        """
        return self._list.S

    def _post_structure_clone(self):
        self._meta['span'] = 1
        self._meta['meas'] = '1/1'
        self._meta['type'] = None
        for node in self._graph.node_indices():
            self._graph[node] = {'proportion': 1}
        self._evaluate()
        subdivs = self._build_subdivisions()
        s = subdivs[1] if isinstance(subdivs, tuple) and len(subdivs) > 1 else (1,)
        self._list = Group((1, s))

    def _cast_subdivs(self, children):
        def convert_to_tuple(item):
            if isinstance(item, RhythmTree):
                return (item.meas.numerator * item.span, item.subdivisions)
            if isinstance(item, (tuple, list)):
                return tuple(convert_to_tuple(x) for x in item)
            return item
        
        return tuple(convert_to_tuple(child) for child in children)

    def _validate_s_form(self, s):
        """Validate S is in valid S-form. Each element is non-zero int or (D, S) tuple.
        S must have at least 2 elements (e.g. (1,) is invalid)."""
        if isinstance(s, int):
            if s == 0:
                raise ValueError(f"S element cannot be zero: {s}")
            return
        if isinstance(s, (tuple, list)):
            if len(s) < 2:
                raise ValueError(f"S must have at least 2 elements, got {s}")
            for elem in s:
                if isinstance(elem, int):
                    if elem == 0:
                        raise ValueError(f"S element cannot be zero: {elem}")
                elif isinstance(elem, (tuple, list)):
                    if len(elem) != 2:
                        raise ValueError(f"(D S) must have exactly 2 elements, got {len(elem)}: {elem}")
                    d, sub = elem
                    if not isinstance(d, int) or d == 0:
                        raise ValueError(f"(D S): D must be non-zero integer, got D={d}")
                    self._validate_s_form(sub)
                else:
                    raise ValueError(f"S element must be non-zero int or (D S) tuple, got {type(elem).__name__}: {elem}")
            return
        raise ValueError(f"S must be tuple or int, got {type(s).__name__}: {s}")

    def _normalize_s_for_subdivide(self, S):
        """Normalize S for subdivide: int -> (1,)*S; tuple -> validate and return.
        S must represent at least 2 subdivisions (e.g. S>1 for int, len(S)>=2 for tuple)."""
        if isinstance(S, int):
            if S <= 1:
                raise ValueError(f"S must be > 1 when int, got {S}")
            return (1,) * S
        if isinstance(S, (tuple, list)):
            self._validate_s_form(S)
            return tuple(S)
        raise ValueError(f"S must be tuple or int, got {type(S).__name__}: {S}")

    def _build_subdivisions(self, root_node=None):
        """
        Build a subdivisions structure from the current graph state.

        Parameters
        ----------
        root_node : int, optional
            The node to start building from. Default is ``self.root``.

        Returns
        -------
        tuple
            Nested tuple structure representing subdivisions.
        """
        if root_node is None:
            root_node = self.root
        
        def get_node_value(node):
            return self[node].get('proportion', 1)
        
        def get_children(node):
            return list(self.successors(node))
        
        return self._build_nested_structure(root_node, get_node_value, get_children)
    
    def _update_group_structure(self):
        """Update the Group structure, preserving D and rebuilding S."""
        if hasattr(self, '_list'):
            new_subdivisions = self._build_subdivisions()
            if isinstance(new_subdivisions, tuple) and len(new_subdivisions) > 1:
                new_s = new_subdivisions[1]
            else:
                new_s = (1,)
            self._list = Group((self._list.D, new_s))

    def _before_post_mutation(self, scope_node=None, op=None):
        for node in self.nodes:
            node_data = self._graph[node]
            if 'proportion' not in node_data:
                node_data['proportion'] = 1

    def _normalize_node_attrs(self, node, attrs, op=None):
        normalized = dict(attrs) if isinstance(attrs, dict) else {}
        if 'label' in normalized:
            raise ValueError("RhythmTree does not accept 'label'; use 'proportion'")

        if 'tied' in normalized and 'proportion' not in normalized:
            current = self[node].get('proportion', 1)
            normalized['proportion'] = float(current) if normalized['tied'] else int(current)

        if 'proportion' in normalized and 'tied' in normalized:
            if normalized['tied']:
                normalized['proportion'] = float(normalized['proportion'])
            else:
                normalized['proportion'] = int(normalized['proportion'])

        return normalized

    def _validate_node_attrs(self, node, attrs, op=None):
        mutable_keys = {'proportion', 'tied'}
        illegal = [k for k in attrs if k not in mutable_keys]
        if illegal:
            raise ValueError(f"Illegal RhythmTree node attribute update: {illegal}")
        if 'metric_duration' in attrs or 'metric_onset' in attrs:
            raise ValueError("metric_duration and metric_onset are derived and cannot be set directly")

    def _resolve_data_update_scope(self, node, changed_keys, op=None):
        if 'proportion' in changed_keys or 'tied' in changed_keys:
            if node == self.root:
                return self.root
            parent = self.parent(node)
            return self.root if parent is None else parent
        return super()._resolve_data_update_scope(node=node, changed_keys=changed_keys, op=op)

    def _after_post_mutation(self, scope_node=None, op=None):
        self._evaluate(scope_node)
    
    @property
    def durations(self):
        """
        The metric durations of all leaf nodes.

        Returns
        -------
        tuple of Fraction
        """
        return tuple(self.nodes[n]['metric_duration'] for n in self.leaf_nodes)
    
    @property
    def onsets(self):
        """
        The metric onsets of all leaf nodes.

        Returns
        -------
        tuple of Fraction
        """
        return tuple(self.nodes[n]['metric_onset'] for n in self.leaf_nodes)
    
    @property
    def info(self):
        """
        A formatted string summarizing the tree's metadata, subdivisions,
        durations, and onsets.

        Returns
        -------
        str
        """
        ordered_meta = {k: self._meta[k] for k in ['span', 'meas', 'type']}
        ordered_meta['depth'] = self.depth
        ordered_meta['k'] = self.k
        meta_str = ' | '.join(f"{k}: {v}" for k, v in ordered_meta.items())
        
        table_data = [
            [str(r) for r in self.durations],
            [str(o) for o in self.onsets]
        ]
        
        duration_onset_table = tabulate(
            table_data,
            headers=[],
            tablefmt='plain'
        )
        
        table_lines = duration_onset_table.split('\n')
        durations_line = f"Durations: {table_lines[0]}"
        onsets_line = f"Onsets:    {table_lines[1]}"
        
        content = [
            meta_str,
            f"Subdivs: {format_subdivisions(self.subdivisions)}",
            onsets_line,
            durations_line
        ]
        
        width = max(len(line) for line in content)
        border = '-' * width
        
        return (
            f"{border}\n"
            f"{content[0]}\n"
            f"{border}\n"
            f"{content[1]}\n"
            f"{border}\n"
            f"{content[2]}\n"
            f"{content[3]}\n"
            f"{border}\n"
        )
    
    # @property
    # def type(self):
    #     if self._meta['type'] is None:
    #         self._meta['type'] = self._set_type()
    #     return self._meta['type']
    
    def _evaluate(self, root_node=None):
        """
        Evaluate the tree to compute metric durations and onsets.

        Single-pass DFS: computes durations and onsets together in one traversal.
        When root_node is provided, evaluates from that subtree (ancestors must
        already have metric_duration). When None, evaluates from root.

        Parameters
        ----------
        root_node : int, optional
            Subtree root to evaluate from. If None, evaluates from tree root.
        """
        if root_node is None:
            root_node = self.root
        self._graph[self.root]['metric_duration'] = self.meas * self.span
        parent_ratio = self.span * self.meas.to_fraction() if root_node == self.root else Fraction(self[root_node]['metric_duration'])

        leaf_onset_acc = [Fraction(0)]

        def _process_child(child, div, parent_ratio, parent_is_negative):
            child_data = self[child]
            s = child_data.get('proportion', 1)
            child_is_tied = isinstance(s, float) or bool(child_data.get('tied', False))
            if 'meta' in child_data:
                s = s * child_data['meta']['span']
            s = int(s) if isinstance(s, float) else s
            if parent_is_negative and s > 0:
                s = -s
            ratio = Fraction(s, div) * parent_ratio
            if s < 0:
                ratio = -abs(ratio)
            self._graph[child]['metric_duration'] = ratio
            self._graph[child]['tied'] = child_is_tied
            self._graph[child]['proportion'] = float(s) if child_is_tied else s
            if self.out_degree(child) > 0:
                _process_subtree(child, ratio)
            else:
                self._graph[child]['metric_onset'] = leaf_onset_acc[0]
                leaf_onset_acc[0] += abs(ratio)

        def _process_subtree(node, parent_ratio):
            node_data = self._graph[node]
            if 'meta' in node_data:
                node_data['proportion'] = node_data.get('proportion', 1) * node_data['meta']['span']
            label = node_data.get('proportion', 1)
            is_tied = isinstance(label, float) or bool(node_data.get('tied', False))
            self._graph[node]['tied'] = is_tied
            label_value = int(label) if is_tied else label
            self._graph[node]['proportion'] = float(label_value) if is_tied else label_value
            children = list(self.successors(node))
            if not children:
                ratio = Fraction(label_value) * parent_ratio
                self._graph[node]['metric_duration'] = ratio
                self._graph[node]['metric_onset'] = leaf_onset_acc[0]
                leaf_onset_acc[0] += abs(ratio)
                return
            div = int(sum(
                abs(self[c].get('proportion', 1)) *
                (self[c]['meta']['span'] if 'meta' in self[c] else 1)
                for c in children
            ))
            node_is_negative = label_value < 0
            for child in children:
                _process_child(child, div, parent_ratio, node_is_negative)
            self._graph[node]['metric_onset'] = self[children[0]]['metric_onset']

        if root_node != self.root:
            for n in self.leaf_nodes:
                if n in self.subtree_leaves(root_node):
                    break
                leaf_onset_acc[0] += abs(Fraction(self[n]['metric_duration']))

        _process_subtree(root_node, parent_ratio)

        if root_node != self.root:
            leaves_after = [n for n in self.leaf_nodes if list(self.leaf_nodes).index(n) > max(list(self.leaf_nodes).index(l) for l in self.subtree_leaves(root_node))]
            for n in leaves_after:
                self._graph[n]['metric_onset'] = leaf_onset_acc[0]
                leaf_onset_acc[0] += abs(Fraction(self[n]['metric_duration']))
            for node in reversed(list(self.topological_sort())):
                if self.out_degree(node) > 0 and node != root_node:
                    if node not in self.descendants(root_node):
                        children = list(self.successors(node))
                        self._graph[node]['metric_onset'] = self[children[0]]['metric_onset']

    def _set_type(self):
        div = sum_proportions(self.subdivisions)
        if bin(div).count('1') != 1 and div != self.meas.numerator:
            return 'complex'
        return 'complex' if measure_complexity(self.subdivisions) else 'simple'

    def __len__(self):
        return len(self.durations)

    def __str__(self):
        return f"RhythmTree(span={self.span}, meas={self.meas}, subdivisions={format_subdivisions(self.subdivisions)})"

    def __repr__(self):
        return self.__str__()
    
    def subtree(self, node, renumber=True):
        """Extract a rhythm subtree rooted at the given node.
        
        The subtree becomes a new RhythmTree with:
        - span = 1
        - meas = metric_duration of the selected node
        - subdivisions = reconstructed from the subtree structure
        
        Parameters
        ----------
        node : int
            The node to use as the root of the subtree
        renumber : bool, optional
            Whether to renumber the nodes in the new tree (default: True)
            
        Returns
        -------
        RhythmTree
            A new RhythmTree representing the subtree
        """
        if node not in self:
            raise ValueError(f"Node {node} not found in tree")

        subdivisions_structure = self._build_subdivisions(node)
        if isinstance(subdivisions_structure, tuple) and len(subdivisions_structure) > 1:
            subdivisions = subdivisions_structure[1]
        else:
            subdivisions = (1,)
        
        node_duration = self[node].get('metric_duration')
        if node_duration is None:
            meas = '1/1'
        else:
            meas = Meas(node_duration)
        
        new_rt = RhythmTree(span=1, meas=meas, subdivisions=subdivisions)
        
        if renumber:
            new_rt.renumber_nodes()
        
        return new_rt
    
    def graft_subtree(self, target_node, subtree, mode='replace'):
        """
        Graft a subtree onto the tree and re-evaluate metric values.

        Parameters
        ----------
        target_node : int
            The node at which to graft.
        subtree : RhythmTree or Tree
            The subtree to graft.
        mode : str, optional
            Grafting mode (``'replace'`` or ``'append'``). Default is
            ``'replace'``.

        Returns
        -------
        RhythmTree
            The modified tree (self).
        """
        return super().graft_subtree(target_node, subtree, mode)

    def subdivide(self, node, S):
        """
        Subdivide leaf node(s) with structure S.

        Each leaf becomes a parent with children defined by S. The node's
        proportion D is used as the duration; S specifies the subdivisions.

        Parameters
        ----------
        node : int or list of int
            The leaf node(s) to subdivide. Must have no children.
            If a list, subdivide is applied to each.
        S : tuple or int
            Subdivisions. If int, interpreted as ``(1,)*S`` (e.g. ``S=3`` → ``(1, 1, 1)``).
            If tuple, must be valid S-form: each element is a non-zero integer
            or a ``(D, S)`` pair.

        Returns
        -------
        RhythmTree
            self (for chaining)

        Raises
        ------
        ValueError
            If node is not found, is not a leaf, or S is invalid.
        """
        nodes = [node] if isinstance(node, int) else list(node)
        for n in nodes:
            if n not in self:
                raise ValueError(f"Node {n} not found in tree")
            if self.out_degree(n) != 0:
                raise ValueError(f"Node {n} must be a leaf")

        S = self._normalize_s_for_subdivide(S)
        S = self._cast_subdivs(S)

        def add_children(parent, children):
            for child in children:
                if isinstance(child, tuple) and len(child) == 2:
                    D, sub = child
                    child_id = self.add_child(parent, proportion=D)
                    add_children(child_id, sub)
                else:
                    self.add_child(parent, proportion=child)

        for n in nodes:
            add_children(n, S)
        scope = self.parent(nodes[0]) if len(nodes) == 1 else None
        self._post_mutation(scope_node=scope, op='subdivide')
        return self

    def prune(self, node):
        super().prune(node)
        return self

    def remove_subtree(self, node):
        super().remove_subtree(node)
        return self

    def make_rest(self, node):
        """
        Make a node and all its descendants into rests by setting their proportions to negative.
        
        Parameters
        ----------
        node : int
            The node ID to make into a rest along with all its descendants.
            
        Raises
        ------
        ValueError
            If the node is not found in the tree.
        """
        if node not in self:
            raise ValueError(f"Node {node} not found in tree")
        
        descendants_to_modify = [node] + list(self.descendants(node))
        
        for n in descendants_to_modify:
            node_data = self._graph[n]
            if 'proportion' in node_data and node_data['proportion'] > 0:
                node_data['proportion'] = -abs(node_data['proportion'])
                node_data['metric_duration'] = -abs(node_data['metric_duration'])
        self._update_group_structure()



import numpy as np

from klotho import (
    CompositionalUnit as UC,
    TemporalUnitSequence as UTS,
    TemporalBlock as BT,
    Pattern,
    Envelope,
)
from klotho.thetos import ToneInstrument as JsInst
from klotho.utils.playback.tonejs.converters import (
    compositional_unit_to_events,
    temporal_sequence_to_events,
    temporal_block_to_events,
)
from klotho.semeios.notelists.supercollider import Scheduler


def test_uc_converter_smoke_with_instruments():
    np.random.seed(7)
    uc = UC(tempus='4/4', prolatio=(1, 1, 1, 1), beat='1/4', bpm=120, inst=JsInst.Kick())
    inst_pat = Pattern([JsInst.Kick(), JsInst.Snare(), JsInst.HatClosed()])
    for leaf in uc.rt.leaf_nodes:
        uc.set_instrument(leaf, next(inst_pat))
        uc.set_pfields(leaf, vel=np.random.uniform(0.5, 0.9))
    payload = compositional_unit_to_events(uc)
    assert len(payload["events"]) == 4
    assert payload["instruments"]


def test_uts_and_bt_converter_smoke():
    uc1 = UC(tempus='10/16', prolatio=((3, (1,) * 4), (4, (1,) * 6), (3, (1,) * 4)), beat='1/16', bpm=140, inst=JsInst.Kick())
    uc2 = UC(tempus='10/16', prolatio=((5, (1,) * 5),) * 2, beat='1/16', bpm=140, inst=JsInst.Snare())
    uts1 = UTS([uc1, uc1.copy()])
    uts2 = UTS([uc2, uc2.copy()])
    bt = BT([uts1, uts2])

    seq_payload = temporal_sequence_to_events(uts1)
    block_payload = temporal_block_to_events(bt)

    assert len(seq_payload["events"]) > 0
    assert len(block_payload["events"]) >= len(seq_payload["events"])


def test_scheduler_handles_uc_slur_and_env_overlay():
    uc = UC(tempus='4/4', prolatio=((2, (1, 1)), (2, (1, 1))), beat='1/4', bpm=120, inst=JsInst.Kalimba())
    inner_nodes = uc.rt.at_depth(1)
    uc.apply_slur(node=inner_nodes[0])
    uc.apply_envelope(envelope=Envelope([0.1, 0.8], times=[1.0]), pfields='vel', node=uc.rt.root)

    scheduler = Scheduler()
    scheduler.add(uc)
    assert scheduler.total_events > 0



'''
--------------------------------------------------------------------------------------

`Tonos` is a specialized module for working with pitch and frequency in music.

In Ancient Greek, "τόνος" (tonos) originally meant "tension," "tone," or "pitch." This 
word has contributed to various terms in modern languages, especially in the fields of 
music, literature, and medicine.

In music, "tonos" is the origin of the word "tone," which is used to describe both a
musical interval and a musical quality. That is, both the phonemenon of pitch and the
perception of "tone-quality" also known as *timbre*.

"Any tone can succeed any other tone, any tone can sound simultaneously with any other
tone or tones, and any group of tones can be followed by any other group of tones, just
as any degree of tension or nuance can occur in any medium under and kind of stress or
duration.  Successful projection will depend upon the contextual and formal conditions
that prevail, and upon the skill and the soul of the composer." 
  — Vincent Persichetti, from "Twentieth-Century Harmony: Creative Aspects and Practice", 
    Chapter One: Intervals

--------------------------------------------------------------------------------------
'''
from . import utils
from . import systems
from . import scales
from . import chords
from . import pitch

combination_product_sets = systems.combination_product_sets
harmonic_trees = systems.harmonic_trees

from .pitch import (
    Pitch,
    PitchCollection,
    PitchCollectionBase,
    RelativePitchCollection,
    AbsolutePitchCollection,
    RootedPitchCollection,
)
from .scales import Scale
from .scales.scale import InstancedScale
from .chords import Chord, Voicing, ChordSequence
from .chords.chord import InstancedChord, InstancedVoicing
from .pitch import Contour
from .systems.combination_product_sets import CombinationProductSet, Hexany, Dekany, Pentadekany, Eikosany, Hebdomekontany, MasterSet
from .systems.harmonic_trees import HarmonicTree
from .systems.harmonic_trees.spectrum import Spectrum
from .systems.tone_lattices import ToneLattice

from .utils.intervals import ratio_to_cents, cents_to_ratio, cents_to_setclass, ratio_to_setclass
from .utils.intervals import split_partial, harmonic_mean, arithmetic_mean, logarithmic_distance
from .utils.intervals import interval_cost, n_tet, ratios_n_tet

from .utils.frequency_conversion import freq_to_midicents, midicents_to_freq
from .utils.frequency_conversion import midicents_to_pitchclass, freq_to_pitchclass, pitchclass_to_freq
from .utils.frequency_conversion import A4_Hz, A4_MIDI, PITCH_CLASSES

from .utils.harmonics import partial_to_fundamental, first_equave

from .utils.interval_normalization import equave_reduce, reduce_interval, reduce_interval_relative
from .utils.interval_normalization import reduce_sequence_relative, fold_interval, reduce_freq

__all__ = [
    # Modules
    'utils',
    'systems',
    'scales',
    'chords',
    'pitch',
    'combination_product_sets',
    'harmonic_trees',
    
    # Pitch Collection Classes
    'Pitch',
    'PitchCollection',
    'PitchCollectionBase',
    'RelativePitchCollection',
    'AbsolutePitchCollection',
    'RootedPitchCollection',
    'Scale',
    'Chord',
    'Voicing',
    
    # Sequences
    'ChordSequence',
    
    # Instanced Pitch Collection Classes
    'InstancedScale',
    'InstancedChord',
    'InstancedVoicing',
    
    # Other Classes
    'Contour',
    'CombinationProductSet',
    'Hexany',
    'Dekany',
    'Pentadekany',
    'Eikosany',
    'Hebdomekontany',
    'HarmonicTree',
    'Spectrum',
    'ToneLattice',
    
    # Interval utilities
    'ratio_to_cents',
    'cents_to_ratio',
    'cents_to_setclass',
    'ratio_to_setclass',
    'split_partial',
    'harmonic_mean',
    'arithmetic_mean',
    'logarithmic_distance',
    'interval_cost',
    'n_tet',
    'ratios_n_tet',
    
    # Frequency conversion utilities
    'freq_to_midicents',
    'midicents_to_freq',
    'midicents_to_pitchclass',
    'freq_to_pitchclass',
    'pitchclass_to_freq',
    'A4_Hz',
    'A4_MIDI',
    'PITCH_CLASSES',
    
    # Harmonics utilities
    'partial_to_fundamental',
    'first_equave',
    
    # Interval normalization utilities
    'equave_reduce',
    'reduce_interval',
    'reduce_interval_relative',
    'reduce_sequence_relative',
    'fold_interval',
    'reduce_freq',
]
